import sys
from pathlib import Path
from importlib import import_module
from unittest.mock import patch

from mkdocs.plugins import BasePlugin
from mkdocs.config import config_options


class Plugin(BasePlugin):
    config_scheme = (("build_locales", config_options.Type(bool, default=False)),)

    def on_config(self, config):
        """Look for and load main_mkdocs_plugin in tools/faker_docs_utils/mkdocs_plugins.py
        This bootstrap plugin is needed because that other one is never "installed"
        It is just present in the repo. So it can't have an official entry point
        in setup.py.
        """
        docs_dir = config["docs_dir"]
        plugins_dir = Path(docs_dir).parent / "tools/faker_docs_utils/mkdocs_plugins"
        new_sys_path = [*sys.path, str(plugins_dir)]
        with patch.object(sys, "path", new_sys_path):
            module = import_module("main_mkdocs_plugin")
            main_plugin = module.Plugin()
            config["plugins"]["main_mkdocs_plugin"] = main_plugin
            main_plugin.config = self.config
            main_plugin.on_config(config)



"""Tools for safely saving and loading pickles using an AllowedList"""

import copyreg
import io
import pickle
import typing as T

import warnings


DispatchDefinition = T.Callable[[T.Any], T.Tuple[type, T.Tuple]]


class RestrictedPickler:
    def __init__(
        self,
        dispatchers: T.Mapping[type, DispatchDefinition],
        _SAFE_CLASSES: T.Set[T.Tuple[str, str]],
    ) -> None:
        self._DISPATCH_TABLE = copyreg.dispatch_table.copy()
        self._DISPATCH_TABLE.update(dispatchers)
        self.RestrictedUnpickler = _get_RestrictedUnpicklerClass(_SAFE_CLASSES)

    def dumps(self, data):
        """Only allow saving "safe" classes"""
        outs = io.BytesIO()
        pickler = pickle.Pickler(outs)
        pickler.dispatch_table = self._DISPATCH_TABLE
        pickler.dump(data)
        return outs.getvalue()

    def loads(self, s):
        """Helper function analogous to pickle.loads()."""
        return self.RestrictedUnpickler(io.BytesIO(s)).load()


class Type_Cannot_Be_Used_With_Random_Reference(T.NamedTuple):
    """This type cannot be unpickled."""

    module: str
    name: str


def _get_RestrictedUnpicklerClass(_SAFE_CLASSES):
    class RestrictedUnpickler(pickle.Unpickler):
        """Safe unpickler with an allowed-list"""

        count = 0

        def find_class(self, module, name):
            # Only allow safe classes from builtins.
            if (module, name) in _SAFE_CLASSES:
                return super().find_class(module, name)
            else:
                # warn first 10 times
                if RestrictedUnpickler.count < 10:
                    warnings.warn(f"Cannot save and refer to {module}, {name}")
                    RestrictedUnpickler.count += 1
                # Return a "safe" object that does nothing.
                return lambda *args: Type_Cannot_Be_Used_With_Random_Reference(
                    module, name
                )

    return RestrictedUnpickler



from snowfakery import SnowfakeryPlugin
from snowfakery.plugins import PluginOption

plugin_options_version = (
    "snowfakery.standard_plugins.SnowfakeryVersion.snowfakery_version"
)


class SnowfakeryVersion(SnowfakeryPlugin):
    allowed_options = [
        PluginOption(plugin_options_version, int),
    ]



import typing as T
import string
from snowfakery.fakedata.fake_data_generator import FakeData

from snowfakery.plugins import PluginContext


class StringGenerator:
    """Sometimes in templates you want a reference to a variable to
    call a function.

    For example:

    >>> x = template_utils.StringGenerator(datetime.today().isoformat)
    >>> print(f"{x}")
    2019-09-23T11:49:01.994453

    >>> x = template_utils.StringGenerator(lambda:str(random.random()))
    >>> print(f"{x}")
    0.795273959965055
    >>> print(f"{x}")
    0.053061903749985206
    """

    def __init__(self, func):
        self.func = func

    def __str__(self):
        return str(self.func())

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def __add__(self, other):
        return str(self) + str(other)

    def __radd__(self, other):
        return str(other) + str(self)

    def render(self):
        return self.func()


class FakerTemplateLibrary:
    """A Jinja template library to add the fake.xyz objects to templates"""

    def __init__(
        self,
        faker_providers: T.Sequence[object],
        locale: T.Optional[str] = None,
        context: T.Optional[PluginContext] = None,
    ):
        self.locale = locale
        self.context = context

        self.fake_data = FakeData(faker_providers, locale, self.context)

    def _get_fake_data(self, name):
        return self.fake_data._get_fake_data(name)

    def __getattr__(self, name):
        return StringGenerator(
            lambda *args, **kwargs: self.fake_data._get_fake_data(name, *args, **kwargs)
        )


number_chars = set(string.digits + ".")


def look_for_number(arg):
    looks_like_float = False
    if len(arg) == 0 or (arg[0] == "0" and arg[1:2] != "."):
        return arg
    for char in arg:
        if char not in number_chars:
            return arg
        if char == ".":
            if looks_like_float:
                # we already saw a ".", so this string must be
                # of the form ###.###.### like a euro-phone #
                return arg
            else:
                looks_like_float = True
    if looks_like_float:
        return float(arg)
    else:
        return int(arg)



import os
import time
from itertools import count
from math import log
import string

from baseconv import BaseConverter


from snowfakery import SnowfakeryPlugin
from snowfakery.plugins import PluginResult, PluginOption
from snowfakery import data_gen_exceptions as exc

from snowfakery.utils.scrambled_numbers import scramble_number

# the option name that the user specifies on the CLI or API is just "pid"
# but using this long name internally prevents us from clashing with the
# user's variable names.
plugin_option_pid = "snowfakery.standard_plugins.UniqueId.UniqueId.pid"
plugin_option_big_ids = "snowfakery.standard_plugins.UniqueId.UniqueId.big_ids"

"""
 == UniqueID Plugin ==

Simplest possible usage is like this:

        - plugin: snowfakery.standard_plugins.UniqueId
        - object: Example
          count: 2
          fields:
            unique: ${{UniqueId.unique_id}}

You will get a roughly 17 character identifier. This identifier
incorporates something called the `current portion identifier` and
a per-element index.

You can create your own, custom unique value generator using a
"generator template":

        - plugin: snowfakery.standard_plugins.UniqueId
        - var: MyGenerator
          value:
            UniqueId.NumericIdGenerator: index
        - object: Example
          count: 2
          fields:
            unique: ${{MyGenerator.unique_id}}

`index` in that example serves as the "template" for the generator.

This particular template will generate very small identifiers because it
does not attempt to generate values that will be unique across
processes running at the same time, or across multiple loads. It just
uses a local index that starts as a single digit and grows.

The default template incorporates something called a "portion identifier"
along with the index. The portion identifier essentially guarantees
uniqueness even if the recipe is run on two different computers
or on the same computer at two different times of day. If you wanted to
include the portion identifer yourself (as the default template does),
you would use this template:

        - plugin: snowfakery.standard_plugins.UniqueId
        - var: MyGenerator1
          value:
            UniqueId.NumericIdGenerator: pid, index

You can also ask the algorithm to "mix in" a number you specify yourself:

        - plugin: snowfakery.standard_plugins.UniqueId
        - var: MyGenerator1
          value:
            UniqueId.NumericIdGenerator: 11, pid, index

Or based on a variable or recipe option:

        - plugin: snowfakery.standard_plugins.UniqueId
        - option: computer_number
        - var: MyGenerator1
          value:
            UniqueId.NumericIdGenerator: ${{computer_number}}, pid, index

"""

### Implementation strategy
#
# Imagine that we want to treat a group of numbers like a hierarchical namespace
# Analogous to subomain.domain.tld or 127.99.0.1 or /directory/directory/directory
#
# In each of the examples above, we use a separator character. But we want our
# numbers to be integers with no obvious separators.
#
# Our trick: convert the numbers to octal and use "9" as the separator.
#
# So given these numbers: [127, 99, 0, 1] we can do this:
#
# >>> lst = [127, 99, 0, 1]
# >>> octnums = [oct(x) for x in lst]
# >>> octnums
# ['0o177', '0o143', '0o0', '0o1']
# >>> as_str = "9".join(octnum[2:] for octnum in octnums)
# >>> as_str
# '17791439091'
#
# This is a unique and reversible representation of that list of numbers.
#
# Note that the number "8" and the string "99" will never appear in these
# numbers, so they are a subset of the normal integers.
#
# TODO: Benchmark Cantor Tuples instead.


def _oct(number):
    return oct(number)[2:]


class UniqueNumericIdGenerator(PluginResult):
    context_uniqifier = count(1)

    def __init__(
        self,
        *,
        parts: str,
        pid: int = None,
        min_chars: int = None,
        randomize: bool = True,
        start: int = 1,
    ):
        self.unique_identifer = next(self.context_uniqifier)
        self.counter = count(start)
        self.start = start
        self.parts = parts
        self.pid = self._get_pid(pid)
        parts = [self._convert(part.strip().lower()) for part in parts.split(",")]
        self.number_template = "9".join(parts)
        self.min_chars = min_chars
        self.result = {}  # implementation detail of PluginResults
        self.randomize = randomize

    def _get_pid(self, pid) -> str:
        if isinstance(pid, int):
            return _oct(pid)
        elif pid is not None:  # pragma: no cover
            assert type(pid) in (
                int,
                type(None),
            ), f"Unsupported datatype for pid: {pid}"

        return (
            _oct(int(time.time() - time.mktime((2021, 1, 1, 0, 0, 0, 0, 0, 0))))
            + "9"
            + _oct(os.getpid())
        )

    def _convert(self, part):
        if isinstance(part, str):
            part = part.lower()
        if part == "pid":
            return self.pid
        # possible future feature: rand8, rand16, etc.
        # elif part.startswith("rand"):
        #     # note that rand is only evaluated once per generator! Not for every generation
        #     numbits = int(part[4:])
        #     return _oct(random.getrandbits(numbits))
        elif part.isnumeric() or isinstance(part, int):
            return _oct(int(part))
        elif part == "index":
            return "{index:o}"
        elif part == "context":
            return _oct(self.unique_identifer)
        else:
            raise exc.DataGenValueError(f"Unknown input to eval: {part}")

    @property
    def unique_id(self) -> int:
        index = next(self.counter)
        val = self.number_template.format(index=index)
        if self.randomize:
            return scramble_number(int(val))
        else:
            return int(val)

    def __reduce__(self):
        state = {
            # don't include pid: continuation processes shoud have their own.
            "parts": self.parts,
            "min_chars": self.min_chars,
            "randomize": self.randomize,
            "start": self.start,
        }
        return (
            self.__class__,
            (state,),
        )


class AlphaUniquifier(PluginResult):
    def __init__(
        self,
        *,
        pid: int = None,
        parts: str = None,
        alphabet: str = None,
        randomize_codes: bool = True,
        min_chars: int = 8,
    ):
        self.randomize_codes = randomize_codes

        # can't randomize extremely small numbers
        if randomize_codes:
            min_chars = max(min_chars, 4)
        self.number_generator = UniqueNumericIdGenerator(
            pid=pid, parts=parts, start=1001, randomize=False
        )
        self.alphabet = alphabet or string.digits + string.ascii_uppercase
        self.alpha_encoder = BaseConverter(self.alphabet).encode
        self.min_chars = min_chars
        self.result = {}  # implementation detail of PluginResults

    def _randomize_number(self, number: int):
        bits_per_char = int(log(len(self.alphabet), 2))
        min_bits = int(self.min_chars) * bits_per_char
        return scramble_number(int(number), min_bits)

    @property
    def unique_id(self) -> str:
        next_number = int(self.number_generator.unique_id)
        if self.randomize_codes:
            next_number = self._randomize_number(next_number)
        return self.alpha_encoder(next_number).rjust(self.min_chars, self.alphabet[0])


def as_bool(opt):
    if isinstance(opt, str) and opt.lower() in ["true", "1", "yes"]:
        return True
    elif isinstance(opt, str) and opt.lower() in ["false", "0", "no"]:
        return False
    elif isinstance(opt, int):
        return bool(opt)
    else:
        raise TypeError(opt)


class UniqueId(SnowfakeryPlugin):
    allowed_options = [
        PluginOption(plugin_option_pid, int),
        PluginOption(plugin_option_big_ids, as_bool),
    ]

    class Functions:
        _default_unique_id_generator = None
        _default_unique_alpha_code_generator = None

        @property
        def _pid(self):
            fieldvars = self.context.field_vars()
            return fieldvars.get(plugin_option_pid)

        @property
        def _bigids(self):
            fieldvars = self.context.field_vars()
            return fieldvars.get(plugin_option_big_ids)

        @property
        def default_uniqifier(self):
            if not self._default_unique_id_generator:
                self._default_unique_id_generator = self.NumericIdGenerator()
            return self._default_unique_id_generator

        @property
        def default_alpha_code_generator(self):
            if not self._default_unique_alpha_code_generator:
                self._default_unique_alpha_code_generator = self.AlphaCodeGenerator()
            return self._default_unique_alpha_code_generator

        @property
        def unique_id(self):
            return self.default_uniqifier.unique_id

        def NumericIdGenerator(self, _=None, *, template: str = None):
            template = template or (
                "pid,context,index" if self._bigids else "context,index"
            )
            return UniqueNumericIdGenerator(pid=self._pid, parts=template)

        def AlphaCodeGenerator(
            self,
            template=None,
            alphabet=None,
            min_chars: int = 8,
            randomize_codes: bool = True,
        ):
            alphabet = str(alphabet) if isinstance(alphabet, int) else alphabet
            template = template or ("pid,context,index" if self._bigids else "index")

            return AlphaUniquifier(
                pid=self._pid,
                parts=template,
                alphabet=alphabet,
                min_chars=min_chars,
                randomize_codes=randomize_codes,
            )



import warnings
from typing import IO, Optional, Tuple, Mapping, List, Dict, TextIO, Union
import typing as T
import functools

import yaml
from faker.providers import BaseProvider as FakerProvider
from click.utils import LazyFile

from snowfakery.standard_plugins.SnowfakeryVersion import SnowfakeryVersion

from .data_gen_exceptions import DataGenNameError
from .output_streams import OutputStream, SimpleFileOutputStream
from .parse_recipe_yaml import parse_recipe
from .data_generator_runtime import (
    Globals,
    Interpreter,
)
from .data_gen_exceptions import DataGenError
from .plugins import SnowfakeryPlugin, PluginOption

from .utils.yaml_utils import SnowfakeryDumper, hydrate
from snowfakery.standard_plugins.UniqueId import UniqueId

# This tool is essentially a three stage interpreter.
#
# 1. Yaml parsing into Python data structures.
# 2. Walking the tree, sorting things into groups like macros, file inclusions,
#    etc., and doing the file inclusions (parse_recipe_yaml.parse_recipe)
# 2 a) merge options informtion from the parse with options from the
#      environment
# 3. Generating the objects top to bottom (including evaluating Jinja) in
#    data_generator_runtime.output_batches
#
# The function generate at the bottom of this file is the entry point to all
# of it.


OpenFileLike = Union[TextIO, LazyFile]


class ExecutionSummary:
    """Summarize everything that happened during parsing and evaluating."""

    def __init__(self, parse_results, runtime_results):
        self.tables = parse_results.tables
        self.templates = parse_results.templates
        self.intertable_dependencies = runtime_results.intertable_dependencies

    def summarize_for_debugging(self):  # pragma: no cover
        return self.intertable_dependencies, self.templates


def merge_options(
    option_definitions: List, user_options: Mapping, raw_plugin_options: Mapping = None
) -> Tuple[Dict, set]:
    """Merge/compare options specified by end-user to those declared in YAML file.

    Takes options passed in from the command line or a config file and
    compare them to the options declared by the Generator YAML file.

    The options from the Generator YAML should be dictionaries with keys of
    "options" and "default" as described in the user documentation.

    The options from the user should be a dictionary of key/value pairs.

    The output is a pair, options, extra_options. The options are the values
    to be fed into the process after applying defaults.

    extra_options are options that the user specified which do not match
    anything in the YAML generator file. The caller may want to warn the
    user about them or throw an error.
    """
    options = raw_plugin_options.copy() if raw_plugin_options else {}
    for option in option_definitions:
        name = option["option"]
        if user_options.get(name):
            options[name] = user_options.get(name)
        elif option.get("default"):
            options[name] = option["default"]
        else:
            raise DataGenNameError(
                f"No definition supplied for option {name}",
            )

    extra_options = set(user_options.keys()) - set(options.keys())
    return options, extra_options


def load_continuation_yaml(continuation_file: OpenFileLike):
    """Load a continuation file from YAML."""
    return hydrate(Globals, yaml.safe_load(continuation_file))


def save_continuation_yaml(continuation_data: Globals, continuation_file: OpenFileLike):
    """Save the global interpreter state from Globals into a continuation_file"""
    yaml.dump(
        continuation_data.__getstate__(),
        continuation_file,
        Dumper=SnowfakeryDumper,
    )


def process_plugins(plugins: List) -> Tuple[List[object], Mapping[str, object]]:
    """Resolve a list of names for SnowfakeryPlugins and Faker Providers to objects

    The Providers are returned as a list of objects.
    The Plugins are a mapping of ClassName:object so they can be namespaced.
    """
    faker_providers = [
        provider for baseclass, provider in plugins if baseclass == FakerProvider
    ]
    snowfakery_plugins = {
        plugin.__name__: plugin
        for baseclass, plugin in plugins
        if baseclass == SnowfakeryPlugin
    }
    return (faker_providers, snowfakery_plugins)


def generate(
    open_yaml_file: IO[str],
    user_options: Optional[dict] = None,
    #  *,   TODO: fix test suite so these can be keyword-only arguments
    output_stream: Optional[OutputStream] = None,
    parent_application=None,
    *,
    stopping_criteria=None,
    generate_continuation_file: OpenFileLike = None,
    continuation_file: TextIO = None,
    plugin_options: dict = None,
    update_input_file: OpenFileLike = None,
    update_passthrough_fields: T.Sequence[str] = (),
) -> ExecutionSummary:
    """The main entry point to the package for Python applications."""
    from .api import SnowfakeryApplication

    user_options = user_options or {}

    # Where are we going to put the rows?
    output_stream = output_stream or SimpleFileOutputStream()

    # parse the YAML and any it refers to
    parse_result = parse_recipe(
        open_yaml_file, update_input_file, update_passthrough_fields
    )

    faker_providers, snowfakery_plugins = process_plugins(parse_result.plugins)

    snowfakery_plugins.setdefault("UniqueId", UniqueId)
    snowfakery_plugins.setdefault("SnowfakeryVersion", SnowfakeryVersion)
    plugin_options = plugin_options or {}
    if parse_result.version:
        plugin_options["snowfakery_version"] = parse_result.version

    plugin_options = process_plugins_options(snowfakery_plugins, plugin_options)

    # figure out how it relates to CLI-supplied generation variables
    options, extra_options = merge_options(
        parse_result.options, user_options, plugin_options
    )

    if extra_options:
        warnings.warn(f"Warning: unknown options: {extra_options}")

    output_stream.create_or_validate_tables(parse_result.tables)

    continuation_data = (
        load_continuation_yaml(continuation_file) if continuation_file else None
    )

    faker_providers, snowfakery_plugins = process_plugins(parse_result.plugins)

    globls = initialize_globals(continuation_data, parse_result.templates)

    # for unit tests that call this function directly
    # they should be updated to use generate_data instead
    parent_application = parent_application or SnowfakeryApplication(stopping_criteria)

    try:
        # now do the output
        with Interpreter(
            output_stream=output_stream,
            options=options,
            snowfakery_plugins=snowfakery_plugins,
            parent_application=parent_application,
            faker_providers=faker_providers,
            parse_result=parse_result,
            globals=globls,
            continuing=bool(continuation_data),
        ) as interpreter:
            runtime_context = interpreter.execute()

    except DataGenError as e:
        if e.filename:
            raise
        else:
            e.filename = getattr(open_yaml_file, "name", None)
            raise

    if generate_continuation_file:
        save_continuation_yaml(runtime_context, generate_continuation_file)

    return ExecutionSummary(parse_result, runtime_context)


def process_plugins_options(
    plugins: Mapping[str, SnowfakeryPlugin], raw_plugin_options: Mapping[str, object]
) -> Mapping[str, object]:
    """Replace option short names with fully qualified names
       and convert types of options.
    e.g. the option name that the user specifies on the CLI or API is just "org_name"
         but we use the long name internally to avoid clashing with the
         user's variable names."""
    allowed_options = collect_allowed_plugin_options(tuple(plugins.values()))
    plugin_options = {}
    for option in allowed_options:
        option_long_name = option.name
        option_short_name = option_long_name.rsplit(".", 1)[-1]
        if option_short_name in raw_plugin_options:
            plugin_options[option_long_name] = option.convert(
                raw_plugin_options[option_short_name]
            )
        elif option_long_name in raw_plugin_options:
            plugin_options[option_long_name] = option.convert(
                raw_plugin_options[option_long_name]
            )
    return plugin_options


def collect_allowed_plugin_options(plugins: Tuple) -> List[PluginOption]:
    """Collect the list of every option allowed by every used plugin"""
    all_option_lists = [getattr(plugin, "allowed_options", []) for plugin in plugins]
    return functools.reduce(lambda x, y: x + y, all_option_lists, [])


def initialize_globals(continuation_data, templates):
    if continuation_data:
        globals = continuation_data
    else:
        name_slots = {
            template.nickname: template.tablename
            for template in templates
            if template.nickname
        }

        tablenames = {template.tablename: template.tablename for template in templates}

        reused_names = set(name_slots).intersection(tablenames)
        if reused_names:
            warnings.warn(
                f"Should not reuse names as both nickname and table name: {reused_names}"
            )
        # table names are sort of nicknames for themselves too, because
        # you can refer to them.
        name_slots.update(tablenames)

        globals = Globals(name_slots=name_slots)

    return globals



import json
from glob import glob
from pathlib import Path

import click
import yaml
from jsonschema import validate


@click.command()
@click.argument("filespecs", nargs=-1)
def validate_recipe(filespecs):
    schemajson = (
        Path(__file__).parent.parent.parent / "schema/snowfakery_recipe.jsonschema.json"
    )
    assert schemajson.exists()

    with schemajson.open() as f:
        schema = json.load(f)
    files = []
    for filespec in filespecs:
        files.extend(glob(filespec))

    if not files:
        raise click.ClickException("No files matched!")

    for file in files:
        with open(file) as f:
            data = yaml.safe_load(f)
            try:
                validate(instance=data, schema=schema)
            except Exception as e:
                print(f"ERROR with file {file}: {e}")


if __name__ == "__main__":
    validate_recipe()



import sys

from typing import Any, Callable, Mapping, Union, NamedTuple, List, Tuple
from importlib import import_module
from datetime import date, datetime
from pathlib import Path
from unittest.mock import patch
from functools import wraps
import typing as T

import yaml
from yaml.representer import Representer
from faker.providers import BaseProvider as FakerProvider
from dateutil.relativedelta import relativedelta

import snowfakery.data_gen_exceptions as exc
from .utils.yaml_utils import SnowfakeryDumper
from .utils.collections import CaseInsensitiveDict

from numbers import Number


Scalar = Union[str, Number, date, datetime, None, relativedelta]
FieldDefinition = "snowfakery.data_generator_runtime_object_model.FieldDefinition"
ObjectRow = "snowfakery.object_rows.ObjectRow"
ScalarTypes = T.get_args(Scalar)


class LineTracker(NamedTuple):
    filename: str
    line_num: int


class SnowfakeryPlugin:
    """Base class for all plugins.

    Currently plugins can declare new functions. Over time new superpowers
    for plugins will be introduced.

    Plugins generally look like this:

    class MyPlugin:
        class Functions: # must be named `Functions`.
            def func1(self, arg1, arg2, arg3=default):
                return something()

    This function would be invoked through the name `MyPlugin.func1()`

    Despite the name, the Functions namespace could also include
    constants or other namespaces, if that is useful to you.

    A RuntimeContext object is supplied on `self.context` with methods
    context.field_vars() and context.context_vars().

    context.field_vars() are the same field variables that can be used in
    templates.

    context.context_vars() is a mutable mapping contaiing
    values that are available to this object template and that of
    all children.

    context_vars are unique to each plugin and plugins do not see
    (or corrupt, or overwrite) context_vars for another plugin.

    Plugins can also keep internal state for global data, like any other
    Python object.
    """

    def __init__(self, interpreter):
        self.interpreter = interpreter
        self.context = PluginContext(self)

    def custom_functions(self, *args, **kwargs):
        """Instantiate, contextualize and return a function library

        Default behaviour is to return self.Functions()."""
        functions = self.Functions()  # type: ignore
        functions.context = self.context
        return functions

    def close(self, *args, **kwargs):
        pass


class ParserMacroPlugin:
    """Abstract base class for plugins that generate code.

    Note that these are currently undocumented and their
    interface is not necessarily stable."""

    pass


class PluginContext:
    "Exposes certain stable internals to plugins"

    def __init__(self, plugin):
        self.plugin = plugin
        self.interpreter = plugin.interpreter

    def field_vars(self):
        return self.interpreter.current_context.field_vars()

    ## TODO: Deprecate this in favour of get_contextual_state
    ##       which has more smarts about name=, parent= etc.
    def context_vars(self):
        return self.interpreter.current_context.context_vars(id(self.plugin))

    def local_vars(self):
        return self.interpreter.current_context.local_vars.setdefault(
            id(self.plugin), {}
        )

    @property
    def unique_context_identifier(self) -> str:
        """An identifier representing a template context that will be
        unique across iterations (but not portion invocations). It
        allows templates that do counting or iteration for a particular
        template context."""
        context_id = self.interpreter.current_context.unique_context_identifier
        return str(context_id)

    def evaluate_raw(self, field_definition):
        """Evaluate the contents of a field definition"""
        return field_definition.render(self.interpreter.current_context)

    def evaluate(self, field_definition):
        """Evaluate the contents of a field definition and simplify to a primitive value."""
        rc = self.evaluate_raw(field_definition)
        if isinstance(rc, ScalarTypes):
            return rc
        elif hasattr(rc, "simplify"):
            return rc.simplify()
        else:
            raise AssertionError(
                f"Cannot simplify {field_definition}. Perhaps should have used evaluate_raw?"
            )

    @property
    def current_filename(self):
        return self.interpreter.current_context.current_template.filename


def lazy(func: Any) -> Callable:
    """A lazy function is one that expects its arguments to be unparsed"""
    func.lazy = True
    return func


def memorable(func: Any) -> Callable:
    """A memorable function is one that is evaluated once per iteration and its result saved"""
    func.memorable = True

    @wraps(func)
    def newfunc(self, *args, **kwargs):
        return evaluate_memorable_function(self.context, func, self, args, kwargs)

    return newfunc


def evaluate_memorable_function(context, func, self, args, kwargs):
    """Memorable functions store state.

    This function fetches that state unless the context has asked us
    to re-start every time.

    For example, a file would be retrieved over and over again, with
    the file pointer advancing, unless the context asked us to re-open it
    every time, in which case the file pointer would always be at the beginning.

    For-loops are the primary example where we want to re-start an iterator
    every time we evaluate.
    """
    if context.interpreter.current_context.recalculate_every_time:
        return func(self, *args, **kwargs)

    # otherwise, do all of the caching magic
    user_key = kwargs.get("name") or (
        context.unique_context_identifier,
        tuple(args),
        tuple(kwargs.items()),
    )
    key = (
        func.__module__,
        func.__name__,
        user_key,
    )
    return context.interpreter.get_contextual_state(
        name=key,
        parent=kwargs.get("parent", None),
        reset_every_iteration=False,
        make_state_func=lambda: func(self, *args, **kwargs),
    )


def resolve_plugins(
    plugin_specs: List[Tuple[str, object]], search_paths: List[Union[str, Path]]
):
    "Resolve a list of plugins and lineinfos"
    with plugin_path(search_paths):
        plugins = []
        for plugin_spec in plugin_specs:
            plugins.extend(resolve_plugin(*plugin_spec))  # type: ignore
        return plugins


def plugin_path(search_paths):
    cwd_plugins = "./plugins"
    user_plugins = Path.home() / ".snowfakery/plugins"
    new_sys_path = [
        *sys.path,
        *(str(p) for p in search_paths),
        str(cwd_plugins),
        str(user_plugins),
    ]
    return patch.object(sys, "path", new_sys_path)


def resolve_plugin(plugin: str, lineinfo) -> object:
    """Resolve a plugin to a class"""
    cls = resolve_plugin_alternatives(plugin)
    if not cls:
        raise exc.DataGenImportError(
            f"Cannot find plugin: {plugin}", lineinfo.filename, lineinfo.line_num
        )

    categories = []

    if issubclass(cls, FakerProvider):
        categories.append((FakerProvider, cls))
    else:
        if issubclass(cls, SnowfakeryPlugin):
            categories.append((SnowfakeryPlugin, cls))
        if issubclass(cls, ParserMacroPlugin):
            categories.append((ParserMacroPlugin, cls))

    if categories:
        return categories
    else:
        raise exc.DataGenTypeError(
            f"{cls} is not a Faker Provider nor Snowfakery Plugin",
            lineinfo.filename,
            lineinfo.line_num,
        )


def resolve_plugin_alternatives(plugin):
    """Interpret the plugin declaration in 2 ways.
    Return it when we find one that matches.

    1. The new way, which allows shorter declarations.
    2. The old way, which required everything to be spelled out.

    In other words:

    - plugin: foo.bar.Baz

    Means.

    1. The new way:

    from foo.bar.Baz import Baz

    2. The old way:

    from foo.bar import Baz
    """
    prefix, class_name = plugin.rsplit(".", 1)
    testnames = [plugin + "." + class_name, plugin]

    for testname in testnames:
        module_name, class_name = testname.rsplit(".", 1)
        try:
            module = import_module(module_name)
            if hasattr(module, class_name):
                assert getattr(module, class_name)
                return getattr(module, class_name)
        except ModuleNotFoundError:
            pass


class PluginResult:
    """`PluginResult` objects expose a namespace that other code can access through dot-notation.

    PluginResults can be initialized with a dict or dict-like object.

    PluginResults are serialized to continuation files as dicts."""

    def __init__(self, result: Mapping):
        self.result = CaseInsensitiveDict(result)

    def __getattr__(self, name):
        # ensures that it won't recurse
        return self.__dict__["result"][name]

    def __reduce__(self):
        return (self.__class__, (dict(self.result),))

    def __repr__(self):
        return f"<{self.__class__} {repr(self.result)}>"

    def __str__(self):
        return str(self.result)

    @classmethod
    def _from_continuation(cls, args):
        return cls(**args)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        _register_for_continuation(cls)


def _register_for_continuation(cls):
    SnowfakeryDumper.add_representer(cls, Representer.represent_object)  # type: ignore
    yaml.SafeLoader.add_constructor(
        f"tag:yaml.org,2002:python/object/apply:{cls.__module__}.{cls.__name__}",
        lambda loader, node: cls._from_continuation(
            loader.construct_mapping(node.value[0])
        ),
    )


class PluginResultIterator(PluginResult):
    closed = False

    def __init__(self, repeat):
        self.repeat = repeat

    def __iter__(self) -> T.Iterator:
        return self

    def __next__(self):
        return self.next()

    def next(self):
        try:
            return self.next_result()
        except StopIteration:
            if self.repeat:
                self.restart()
                return self.next_result()
            else:
                raise

    def next_result(self):
        "Initialize the iterator in self.results."
        raise NotImplementedError(f"start method on {self.__class__.__name__}")

    def start(self):
        "Initialize the iterator in self.results."
        raise NotImplementedError(f"start method on {self.__class__.__name__}")

    def restart(self):
        "Restart the iterator by assigning to self.results"
        self.start()

    def close(self):
        "Subclasses should implement this if they need to clean up resources"
        pass

    def __del__(self):
        if not self.closed:
            self.close()
            self.closed = True


class PluginOption:
    def __init__(self, name, typ):
        self.name = name
        self.type = typ

    def convert(self, value):
        try:
            return self.type(value)
        except (TypeError, ValueError) as e:
            raise exc.DataGenTypeError(
                f"{self.name} option is wrong type {type(value)} rather than {self.type}",
                *e.args,
            )


# round-trip PluginResult objects through continuation YAML if needed.
_register_for_continuation(PluginResult)



import typing as T
from collections import defaultdict
from datetime import date
from pathlib import Path

import yaml
from pydantic import BaseModel, ConfigDict, PositiveInt, RootModel, field_validator


class AtomicDecl(T.NamedTuple):
    """A single object/key/value declaration

    Might be extracted from a larger declaration set.

    e.g.

    - sf_object: foo
      a: b
      c: d

    Implies:
    AtomicDecl("foo", "a", "b", ...)
    AtomicDecl("foo", "c", "d", ...)
    """

    sf_object: str
    key: str
    value: T.Union[T.List, str, date, int]
    priority: int
    merge_rule: T.Callable  # what to do if two declarations for same val


class MergeRules:
    """Namespace for merge rules.

    Merge rules say what to do if there are two rules
    that apply to the same sobject with the same key."""

    @staticmethod
    def use_highest_priority(
        new_decl: T.Optional[AtomicDecl], existing_decl: T.Optional[AtomicDecl]
    ):
        """The Highlander strategy. There can be only one."""
        if existing_decl:
            return max(existing_decl, new_decl, key=lambda decl: decl.priority)
        return new_decl

    @staticmethod
    def append(new_decl: AtomicDecl, existing_decl: T.Optional[AtomicDecl]):
        """The collaborative strategy. Let's work together."""
        if existing_decl:
            existing_decl.value.append(new_decl.value)
            return existing_decl
        else:
            d = new_decl
            # start to build a list-based declaration
            return AtomicDecl(d.sf_object, d.key, [d.value], d.priority, d.merge_rule)


class SObjectRuleDeclaration(BaseModel):
    model_config = ConfigDict(extra="forbid")

    sf_object: str
    priority: T.Optional[T.Literal["low", "medium", "high"]] = None

    api: T.Optional[T.Literal["smart", "rest", "bulk"]] = None
    batch_size: T.Optional[int] = None
    bulk_mode: T.Optional[T.Literal["serial", "parallel"]] = None
    anchor_date: T.Optional[T.Union[str, date]] = None

    load_after: T.Optional[str] = None

    @property
    def priority_number(self):
        values = {"low": 1, "medium": 2, "high": 3, None: 2}
        return values[self.priority]

    @field_validator("priority", "api", "bulk_mode", mode="before")
    def case_normalizer(cls, val):
        if hasattr(val, "lower"):
            return val.lower()
        else:
            return val

    def as_mapping(self):
        rc = {
            "api": self.api,
            "bulk_mode": self.bulk_mode,
            "batch_size": self.batch_size,
            "anchor_date": self.anchor_date,
        }
        return {k: v for k, v in rc.items() if v is not None}


MERGE_RULES = {
    "api": MergeRules.use_highest_priority,
    "bulk_mode": MergeRules.use_highest_priority,
    "batch_size": MergeRules.use_highest_priority,
    "anchor_date": MergeRules.use_highest_priority,
    "load_after": MergeRules.append,
}


class ChannelDeclaration(BaseModel):
    "Channel declarations are only of relevance to Salesforce employees"
    model_config = ConfigDict(extra="forbid")

    user: str
    recipe_options: T.Optional[T.Dict[str, T.Any]] = None
    num_generators: T.Optional[PositiveInt] = None
    num_loaders: T.Optional[PositiveInt] = None


class ChannelDeclarationList(BaseModel):
    "Channel declarations are only of relevance to Salesforce employees"
    user_channels: T.List[ChannelDeclaration]


class LoadDeclarationsTuple(T.NamedTuple):
    sobject_declarations: T.List[SObjectRuleDeclaration]
    channel_declarations: T.List[
        ChannelDeclaration
    ]  # Channel declarations are only of relevance to Salesforce employees


class SObjectRuleDeclarationFile(
    RootModel[T.List[T.Union[ChannelDeclarationList, SObjectRuleDeclaration]]]
):
    @classmethod
    def parse_from_yaml(cls, f: T.Union[Path, T.TextIO]):
        "Parse from a file-like or Path"
        if isinstance(f, Path):
            with open(f) as fd:
                data = yaml.safe_load(fd)
        else:
            data = yaml.safe_load(f)

        parsed = cls.model_validate(data)

        sobject_decls = [
            obj for obj in parsed.root if isinstance(obj, SObjectRuleDeclaration)
        ]
        channel_decls = [
            obj for obj in parsed.root if isinstance(obj, ChannelDeclarationList)
        ]
        if len(channel_decls) > 1:
            raise AssertionError("Only one channel declaration list allowed per file.")
        elif len(channel_decls) == 1:
            channels = channel_decls[0].user_channels
        else:
            channels = []

        return LoadDeclarationsTuple(sobject_decls, channels)


def atomize_decls(decls: T.Sequence[SObjectRuleDeclaration]):
    rc = []
    for decl in decls:
        for key, merge_rule in MERGE_RULES.items():
            val = getattr(decl, key)
            if val is not None:
                rc.append(
                    AtomicDecl(
                        decl.sf_object, key, val, decl.priority_number, merge_rule
                    )
                )
    return rc


def merge_declarations(decls: T.Sequence[SObjectRuleDeclaration]):
    atomic_decls = atomize_decls(decls)
    results = defaultdict(dict)
    for decl in atomic_decls:
        sobject, key, value, priority_number, merge_rule = decl
        previous_decl = results[decl.sf_object].get(decl.key)
        results[sobject][key] = merge_rule(decl, previous_decl)
    return results


def unify(decls: T.Sequence[SObjectRuleDeclaration]):
    decls = merge_declarations(decls)
    objs = {}
    for sobj, sobjdecls in decls.items():
        unified_decls = objs[sobj] = SObjectRuleDeclaration(sf_object=sobj)
        for key, decl in sobjdecls.items():
            assert decl.key == key
            setattr(unified_decls, key, decl.value)
    return objs



from snowfakery import generate_data
from io import StringIO
from timeit import timeit

yaml = "- object: Foo"


def gen():
    with StringIO(yaml) as f:
        generate_data(f, output_file="/dev/null", output_format="txt")


print(timeit(gen, number=1))



"""Runtime objects and algorithms used during the generation of rows."""
import os
from collections import defaultdict, ChainMap
from datetime import date, datetime, timezone
from contextlib import contextmanager

from typing import Optional, Dict, Sequence, Mapping, NamedTuple, Set
import typing as T
from warnings import warn

import jinja2
from jinja2 import nativetypes
import yaml

from .utils.template_utils import FakerTemplateLibrary
from .utils.yaml_utils import SnowfakeryDumper, hydrate
from .row_history import RowHistory
from .template_funcs import StandardFuncs
from .data_gen_exceptions import DataGenSyntaxError, DataGenNameError
import snowfakery  # noQA
from snowfakery.object_rows import (
    NicknameSlot,
    SlotState,
    ObjectRow,
    ObjectReference,
    RowHistoryCV,
)
from snowfakery.plugins import PluginContext, SnowfakeryPlugin, ScalarTypes
from snowfakery.utils.collections import OrderedSet

OutputStream = "snowfakery.output_streams.OutputStream"
VariableDefinition = "snowfakery.data_generator_runtime_object_model.VariableDefinition"
ObjectTemplate = "snowfakery.data_generator_runtime_object_model.ObjectTemplate"
Statement = "snowfakery.data_generator_runtime_object_model.Statement"
ParseResult = "snowfakery.parse_recipe_yaml.ParseResult"


# save every single object to history. Useful for testing saving of datatypes
SAVE_EVERYTHING = os.environ.get("SF_SAVE_EVERYTHING")


class StoppingCriteria(NamedTuple):
    """When have we iterated over the Snowfakery script enough times?"""

    tablename: str
    count: int


class IdManager:
    """Keep track of the most recent ID per Object type"""

    def __init__(self):
        self.last_used_ids = defaultdict(lambda: 0)
        self.start_ids = {}

    def generate_id(self, table_name: str) -> int:
        self.last_used_ids[table_name] += 1
        return self.last_used_ids[table_name]

    def __getitem__(self, table_name: str) -> int:
        return self.last_used_ids[table_name]

    def __getstate__(self):
        return {"last_used_ids": dict(self.last_used_ids)}

    def __setstate__(self, state):
        self.last_used_ids = defaultdict(lambda: 0, state["last_used_ids"])
        self.start_ids = {name: val + 1 for name, val in self.last_used_ids.items()}


SnowfakeryDumper.add_representer(defaultdict, SnowfakeryDumper.represent_dict)


class Dependency(NamedTuple):
    table_name_from: str
    table_name_to: str
    field_name: str


yaml.SafeDumper.add_representer(
    Dependency, lambda representer, obj: representer.represent_list(obj)
)


class Transients:
    def __init__(self, nicknames_and_tables: Mapping[str, str], id_manager: IdManager):
        self.nicknamed_objects = {}
        self.last_seen_obj_by_table = {}
        self.named_slots = {
            name: NicknameSlot(table, id_manager)
            for name, table in nicknames_and_tables.items()
        }

        self.orig_used_ids = id_manager.last_used_ids.copy()


class Globals:
    """Globally named objects and other aspects of global scope

    This object is designed to be persisted to allow long-running
    Snowfakery executions to stop and restart. Other Interpreter internals
    do not persist. For example, it isn't possible to persist a database
    handle in an output_stream."""

    persistent_nicknames: Dict[str, ObjectRow]  # nicknamed objects
    persistent_objects_by_table: Dict[
        str, ObjectRow
    ]  # objects referencable by table name
    id_manager: IdManager  # keeps track of used ids
    intertable_dependencies: Set[Dependency]  # all intertable dependencies detected
    today: date  # today's date
    nicknames_and_tables: Mapping[str, str]  # what table does each nickname refer to?

    def __init__(
        self,
        today: date = None,
        name_slots: Mapping[str, str] = None,
    ):
        # these lists start empty and are filled.
        # They survive iterations and continuations.
        self.persistent_nicknames = {}
        self.persistent_objects_by_table = {}

        self.id_manager = IdManager()
        self.intertable_dependencies = OrderedSet()
        self.today = today or date.today()
        self.nicknames_and_tables = name_slots or {}

        self.reset_slots()

    def register_object(
        self, obj: ObjectRow, nickname: Optional[str], persistent_object: bool
    ):
        """Register an object for lookup by object type and (optionally) Nickname"""
        if nickname:
            if persistent_object:
                self.persistent_nicknames[nickname] = obj
            else:
                self.transients.nicknamed_objects[nickname] = obj
        if persistent_object:
            self.persistent_objects_by_table[obj._tablename] = obj
        self.transients.last_seen_obj_by_table[obj._tablename] = obj

    @property
    def object_names(self):
        """The globally named objects"""
        # the order is important: later overrides earlier
        # i.e. fulfilled names override "slots"
        return {
            **self.transients.named_slots,  # potential forward or backwards references
            **self.persistent_nicknames,  # long-lived nicknames
            **self.persistent_objects_by_table,  # long-lived objects
            **self.transients.nicknamed_objects,  # local nicknames that have been fulfilled
            **self.transients.last_seen_obj_by_table,  # local tablenames that have been fulfilled
        }

    def generate_id_for_nickname(self, nickname: str):
        slot = self.transients.named_slots.get(nickname)
        if slot and slot.status == SlotState.ALLOCATED:
            return slot.consume_slot()

    def register_intertable_reference(
        self, table_name_from: str, table_name_to: str, fieldname: str
    ):
        self.intertable_dependencies.add(
            Dependency(table_name_from, table_name_to, fieldname)
        )

    def reset_slots(self):
        "At the beginning of every iteration, reset the forward reference slots"
        self.transients = Transients(self.nicknames_and_tables, self.id_manager)

    def check_slots_filled(self):
        not_filled = [
            name
            for name, slot in self.transients.named_slots.items()
            if slot.status == SlotState.ALLOCATED
        ]
        if not_filled:
            plural = "s" if len(not_filled) > 1 else ""
            raise DataGenNameError(
                f"Reference{plural} not fulfilled: {','.join(not_filled)}"
            )

    def __getstate__(self):
        def serialize_dict_of_object_rows(dct):
            return {k: v.__getstate__() for k, v in dct.items()}

        persistent_nicknames = serialize_dict_of_object_rows(self.persistent_nicknames)
        persistent_objects_by_table = serialize_dict_of_object_rows(
            self.persistent_objects_by_table
        )
        intertable_dependencies = [
            dict(v._asdict()) for v in self.intertable_dependencies
        ]  # converts ordered-dict to dict for Python 3.6 and 3.7

        state = {
            "persistent_nicknames": persistent_nicknames,
            "persistent_objects_by_table": persistent_objects_by_table,
            "id_manager": self.id_manager.__getstate__(),
            "today": self.today,
            "nicknames_and_tables": self.nicknames_and_tables,
            "intertable_dependencies": intertable_dependencies,
        }
        return state

    def __setstate__(self, state):
        def deserialize_dict_of_object_rows(dct):
            return {k: hydrate(ObjectRow, v) for k, v in dct.items()}

        self.nicknamed_objects = deserialize_dict_of_object_rows(
            state.get("nicknamed_objects", {})
        )
        self.persistent_nicknames = deserialize_dict_of_object_rows(
            state.get("persistent_nicknames", {})
        )
        self.nicknames_and_tables = state["nicknames_and_tables"]
        self.id_manager = hydrate(IdManager, state["id_manager"])

        self.intertable_dependencies = OrderedSet(
            Dependency(*dep) for dep in getattr(state, "intertable_dependencies", [])
        )

        self.today = state["today"]
        persistent_objects_by_table = state.get("persistent_objects_by_table")
        self.persistent_objects_by_table = (
            deserialize_dict_of_object_rows(persistent_objects_by_table)
            if persistent_objects_by_table
            else {}
        )
        self.reset_slots()


class JinjaTemplateEvaluatorFactory:
    def __init__(self, native_types: bool):
        if native_types:
            self.compilers = [
                nativetypes.NativeEnvironment(
                    block_start_string="${%",
                    block_end_string="%}",
                    variable_start_string="${{",
                    variable_end_string="}}",
                )
            ]

            return

        # TODO: Delete this old code_path when the
        #       transition to native_types is complete.
        self.compilers = [
            jinja2.Environment(
                block_start_string="${%",
                block_end_string="%}",
                variable_start_string="${{",
                variable_end_string="}}",
            ),
            jinja2.Environment(
                block_start_string="<%",
                block_end_string="%>",
                variable_start_string="<<",
                variable_end_string=">>",
            ),
        ]

    def compiler_for_string(self, definition: str):
        for compiler in self.compilers:
            for delimiter in ["block_start_string", "variable_start_string"]:
                if getattr(compiler, delimiter) in definition:
                    return compiler
        return None

    def get_evaluator(self, definition: str):
        assert isinstance(definition, str), definition
        compiler = self.compiler_for_string(definition)

        if compiler:
            try:
                template = compiler.from_string(definition)
                return lambda context: template.render(context.field_vars())
            except jinja2.exceptions.TemplateSyntaxError as e:
                raise DataGenSyntaxError(str(e)) from e
        else:
            return lambda context: definition


class Interpreter:
    """Snowfakery runtime interpreter state."""

    current_context: "RuntimeContext" = None
    iteration_count = 0

    def __init__(
        self,
        output_stream: OutputStream,
        parse_result: ParseResult,
        globals: Globals,
        *,
        parent_application,
        options: Mapping = None,
        snowfakery_plugins: Optional[Mapping[str, callable]] = None,
        faker_providers: Sequence[object] = (),
        continuing=False,
    ):
        self.output_stream = output_stream
        self.options = options or {}
        self.faker_providers = faker_providers
        snowfakery_plugins = snowfakery_plugins or {}
        self.plugin_instances = {
            name: plugin(self) for name, plugin in snowfakery_plugins.items()
        }
        self.plugin_function_libraries = {
            name: plugin.custom_functions()
            for name, plugin in self.plugin_instances.items()
        }
        self.globals = globals
        self.continuing = continuing
        stop_table_name = parent_application.stopping_tablename
        if stop_table_name and stop_table_name not in parse_result.tables:
            raise DataGenNameError(
                f"No template creating {stop_table_name}",
            )

        # make a plugin context for our Faker stuff to act like a plugin
        self.faker_plugin_context = PluginContext(SnowfakeryPlugin(self))

        self.faker_template_libraries = {}

        # inject context into the standard functions
        standard_funcs_obj = StandardFuncs(self).custom_functions()
        self.standard_funcs = {
            name: getattr(standard_funcs_obj, name)
            for name in dir(standard_funcs_obj)
            if not name.startswith("_") and name != "context"
        }

        self.statements = parse_result.statements
        self.parent_application = parent_application
        self.instance_states = {}
        self.filter_row_values = self.filter_row_values_normal
        snowfakery_version = self.options.get(
            "snowfakery.standard_plugins.SnowfakeryVersion.snowfakery_version", 2
        )
        assert snowfakery_version in (2, 3)
        self.native_types = snowfakery_version == 3
        self.template_evaluator_factory = JinjaTemplateEvaluatorFactory(
            self.native_types
        )
        self.tables_to_keep_history_for = find_tables_to_keep_history_for(
            parse_result, globals.nicknames_and_tables
        )
        self.row_history = RowHistory(
            globals.transients.orig_used_ids,
            self.tables_to_keep_history_for,
            self.globals.nicknames_and_tables,
        )
        self.resave_objects_from_continuation(globals, self.tables_to_keep_history_for)

    def resave_objects_from_continuation(
        self, globals: Globals, tables_to_keep_history_for: T.Iterable[str]
    ):
        """Re-save just_once objects to the local history cache after resuming a continuation"""

        # deal with objs known by their nicknames
        relevant_objs = [
            (obj._tablename, nickname, obj)
            for nickname, obj in globals.persistent_nicknames.items()
        ]
        already_saved = set(obj._id for (_, _, obj) in relevant_objs)
        # and those known by their tablename, if not already in the list
        relevant_objs.extend(
            (tablename, None, obj)
            for tablename, obj in globals.persistent_objects_by_table.items()
            if obj._id not in already_saved
        )
        # filter out those in tables that are not history-backed
        relevant_objs = (
            (table, nick, obj)
            for (table, nick, obj) in relevant_objs
            if table in tables_to_keep_history_for
        )
        for tablename, nickname, obj in relevant_objs:
            self.row_history.save_row(tablename, nickname, obj._values)

    def execute(self):
        RowHistoryCV.set(self.row_history)
        self.current_context = RuntimeContext(interpreter=self)
        self.loop_over_templates_until_finished(self.continuing)
        return self.globals

    def faker_template_library(self, locale):
        """Create a faker template library for locale, or retrieve it from a cache"""
        rc = self.faker_template_libraries.get(locale)
        if not rc:
            rc = FakerTemplateLibrary(
                self.faker_providers,
                locale,
                self.faker_plugin_context,
            )
            self.faker_template_libraries[locale] = rc
        return rc

    def loop_over_templates_until_finished(self, continuing):
        finished = False
        self.current_context = RuntimeContext(interpreter=self)
        while not finished:
            self.loop_over_templates_once(self.statements, continuing)
            finished = self.current_context.check_if_finished()
            self.iteration_count += 1
            continuing = True
            self.globals.reset_slots()
            self.row_history.reset_locals()

    def loop_over_templates_once(self, statement_list, continuing: bool):
        for statement in statement_list:
            statement.execute(self, self.current_context, continuing)

    def register_variable(self, name: str, value: object):
        vardefs = self.current_context.variable_definitions()
        vardefs[name] = value

    def __enter__(self, *args):
        return self

    def __exit__(self, *args):
        for plugin in self.plugin_instances.values():
            try:
                plugin.close()
            except Exception as e:
                warn(f"Could not close {plugin} because {repr(e)}")
        self.current_context = None
        self.plugin_instances = None
        self.plugin_function_libraries = None
        self.instance_states = None

    def get_contextual_state(
        self,
        *,
        make_state_func: T.Callable,
        name: T.Union[str, tuple, None] = None,
        parent: T.Optional[str] = None,
        reset_every_iteration: bool = False,
    ):
        """Get state that is specific to a particular template&plugin

        The first time the template is invoked in a particular context,
        make_state_func is invoked to generate the state container.

        The next time it is invoked, the state will be returned for reuse
        and potential modification.

        `name` allows you to reuse the same state among multiple templates.
        The function should generally expose this to the end-user
        through an argument called `name`.

        `parent` allows the user to use some specific Object parent as
        a parent object. The state will be only reused for the lifetime
        of the parent and then discarded. This should also be
        user-controlled.

        `reset_every_iteration` is an experimental feature that should
        not generally be used.
        """
        assert not reset_every_iteration
        current_context = self.current_context
        uniq_name = name or current_context.unique_context_identifier
        if parent:
            parent_obj = current_context.field_vars().get(parent)
        # elif reset_every_iteration:           # in case we bring back this feature
        #     parent_obj = self.iteration_count
        else:
            parent_obj = None
        current_parent, value = self.instance_states.get(uniq_name, (None, None))
        if current_parent != parent_obj or value is None:
            value = make_state_func()
            self.instance_states[uniq_name] = [parent_obj, value]
        return value

    def filter_row_values_normal(self, row: dict):
        return {k: v for k, v in row.items() if not k.startswith("__")}

    # for future use:

    # def filter_row_values_and_remove_ids(self, row: dict):
    #     return {k: v for k, v in row.items() if not k.startswith("__") or k == "id"}


class RuntimeContext:
    """Local "stack frame" type object. RuntimeContexts live on the Python stack.

    There are many proxy methods for other objects which helps keep the
    internals of the runtime hidden from outside code to make maintenance
    easier. From the point of view of other modules this is "the interpreter"
    but internally its mostly just proxying to other classes."""

    obj: Optional[ObjectRow] = None
    current_template = None
    local_vars = None
    unique_context_identifier = None
    recalculate_every_time = False  # by default, data is recalculated constantly

    def __init__(
        self,
        interpreter: Interpreter,
        current_template: Statement = None,
        parent_context: "RuntimeContext" = None,
    ):
        if current_template:
            self.current_table_name = current_template.tablename
            self.current_template = current_template

        self.interpreter = interpreter
        self.parent = parent_context
        if self.parent:
            self._plugin_context_vars = self.parent._plugin_context_vars.new_child()
            # are we in a re-calculate everything context?
            self.recalculate_every_time = parent_context.recalculate_every_time
        else:
            self._plugin_context_vars = ChainMap()
        locale = self.variable_definitions().get("snowfakery_locale")
        self.faker_template_library = self.interpreter.faker_template_library(locale)
        self.local_vars = {}

    @property
    def filter_row_values(self):
        return self.interpreter.filter_row_values

    # TODO: move this into the interpreter object
    def check_if_finished(self):
        "Have we iterated over the script enough times?"
        # check that every forward reference was resolved
        globls = self.interpreter.globals
        app = self.interpreter.parent_application

        globls.check_slots_filled()
        app.ensure_progress_was_made(globls.id_manager)

        return app.check_if_finished(globls.id_manager)

    def generate_id(self, nickname: Optional[str]):
        "Generate a unique ID for this object"
        rc = None
        # check if an ID has already been assigned based on the nickname
        # (due to a forward reference)
        if nickname:
            rc = self.interpreter.globals.generate_id_for_nickname(nickname)
        # check if an ID has already been assigned based on the tablenmae.
        # (due to a forward reference)
        rc = rc or self.interpreter.globals.generate_id_for_nickname(
            self.current_table_name
        )
        # otherwise just create a new one
        rc = rc or self.interpreter.globals.id_manager.generate_id(
            self.current_table_name
        )
        return rc

    def remember_row(self, tablename: str, nickname: T.Optional[str], row: dict):
        for fieldname, fieldvalue in row.items():
            if isinstance(fieldvalue, (ObjectRow, ObjectReference)):
                self.interpreter.globals.register_intertable_reference(
                    tablename, fieldvalue._tablename, fieldname
                )
        history_tables = self.interpreter.tables_to_keep_history_for
        should_save: bool = (
            (tablename in history_tables)
            or (nickname in history_tables)
            or SAVE_EVERYTHING
        )
        if should_save:
            self.interpreter.row_history.save_row(tablename, nickname, row)

    def register_object(self, obj, name: Optional[str], persistent: bool):
        "Keep track of this object in case other objects refer to it."
        self.obj = obj
        self.interpreter.globals.register_object(obj, name, persistent)

    @contextmanager
    def child_context(self, template):
        "Create a nested RuntimeContext (analogous to a 'stack frame')."
        jr = self.__class__(
            current_template=template,
            interpreter=self.interpreter,
            parent_context=self,
        )
        try:
            self.interpreter.current_context = jr
            yield jr
        finally:
            # Goodbye junior, its been nice
            # Hope you find your paradise
            self.interpreter.current_context = self

    @property
    def output_stream(self):
        return self.interpreter.output_stream

    def get_evaluator(self, definition: str):
        return self.interpreter.template_evaluator_factory.get_evaluator(definition)

    @property
    def evaluation_namespace(self):
        return EvaluationNamespace(self)

    def executable_blocks(self):
        return self.evaluation_namespace.executable_blocks()

    def field_vars(self):
        return self.evaluation_namespace.field_vars()

    def context_vars(self, plugin_namespace):
        """Variables which are inherited by child scopes"""
        # This looks like a candidate for optimization.
        # An unconditional object copy seems expensive.
        local_plugin_vars = self._plugin_context_vars.get(plugin_namespace, {}).copy()
        self._plugin_context_vars[plugin_namespace] = local_plugin_vars
        return local_plugin_vars

    def variable_definitions(self):
        return self.context_vars("variable definitions")


# NamedTuple because it is immutable, efficient and auto-generates init
class EvaluationNamespace(NamedTuple):
    """Supplies names for evaluation of YAML trees and Jinja expressions."""

    runtime_context: RuntimeContext

    def simple_field_vars(self):
        "Variables that can be inserted into templates"
        # obj=None in some contexts, e.g. evaluating count
        obj = self.runtime_context.obj
        interpreter = self.runtime_context.interpreter
        return {
            "id": obj.id if obj else None,
            "count": obj.id if obj else None,
            "child_index": obj._child_index if obj else None,
            "this": obj,
            "today": interpreter.globals.today,
            "now": datetime.now(timezone.utc),
            "fake": self.runtime_context.faker_template_library,
            "template": self.runtime_context.current_template,
            **interpreter.options,
            **interpreter.globals.object_names,
            **(obj._values if obj else {}),
            **interpreter.plugin_function_libraries,
            **self.runtime_context.variable_definitions(),
        }

    def field_funcs(self):
        "Injects context into functions from template_funcs module."

        return self.runtime_context.interpreter.standard_funcs

    def executable_blocks(self):
        "Return mapping of functions that can be used in YAML block functions"
        return {**self.field_funcs(), "fake": self.fake}

    def fake(self, name):
        val = self.runtime_context.faker_template_library._get_fake_data(name)
        if isinstance(val, ScalarTypes):
            return val
        else:
            return str(val)

    def field_vars(self):
        return {**self.simple_field_vars(), **self.field_funcs()}


def evaluate_function(func, args: Sequence, kwargs: Mapping, context):
    if not hasattr(func, "lazy"):
        args = [arg.render(context) if hasattr(arg, "render") else arg for arg in args]
        kwargs = {
            name: arg.render(context) if hasattr(arg, "render") else arg
            for name, arg in kwargs.items()
        }
    return func(*args, **kwargs)


def find_tables_to_keep_history_for(
    parse_result: ParseResult, nicknames_and_tables: T.Dict[str, str]
) -> T.Set[str]:
    """Only keep history for certain tables that are actually referred to by random_reference"""
    random_references = parse_result.random_references
    referenced_names = set(
        get_referent_name(random_reference) for random_reference in random_references
    )
    # normalize nicknames to their underlying table
    referenced_tables = set(
        nicknames_and_tables.get(name, name) for name in referenced_names
    )
    return referenced_tables


def get_referent_name(random_reference):
    """What does this random_reference refer to?"""
    args, kwargs = random_reference.args, random_reference.kwargs
    assert not (args and kwargs)
    if args:
        ret = args[0].definition
    elif kwargs:
        ret = kwargs["to"].definition
    if not isinstance(ret, str):
        raise DataGenSyntaxError(
            f"random_reference should only refer to a name, not {ret}"
        )
    return ret



import typing as T

from sqlalchemy import create_engine, MetaData, Column, Table, Unicode
from sqlalchemy.sql import select
from sqlalchemy.engine.base import Connection
from snowfakery.data_gen_exceptions import DataGenError


def create_cci_record_type_tables(db_url: str):
    """Create record type tables that CCI expects"""
    engine = create_engine(db_url)
    metadata = MetaData()
    metadata.reflect(views=True, bind=engine)
    with engine.connect() as connection, connection.begin():  # type: ignore
        for tablename, table in list(metadata.tables.items()):  # type: ignore
            record_type_column = find_record_type_column(
                tablename, table.columns.keys()
            )
            if record_type_column:
                rt_table = _create_record_type_table(tablename, metadata)
                rt_table.create(bind=connection)
                _populate_rt_table(connection, table, record_type_column, rt_table)


def _create_record_type_table(tablename: str, metadata: MetaData) -> Table:
    """Create a table to store mapping between Record Type Ids and Developer Names."""
    rt_map_fields = [
        Column("record_type_id", Unicode(18), primary_key=True),
        Column("developer_name", Unicode(255)),
    ]
    rt_map_table = Table(tablename + "_rt_mapping", metadata, *rt_map_fields)
    return rt_map_table


def _populate_rt_table(
    connection: Connection, table: Table, columnname: str, rt_table: Table
):
    column = getattr(table.columns, columnname)
    query_res = connection.execute(select(column).where(column is not None).distinct())
    record_types = [res[0] for res in query_res if res[0]]  # type: ignore

    if record_types:
        insert_stmt = rt_table.insert()
        rows = [
            dict(zip(rt_table.columns.keys(), (rtname, rtname)))  # type: ignore
            for rtname in record_types
        ]

        connection.execute(insert_stmt, rows)


def find_record_type_column(tablename: str, columnnames: T.Iterable[str]):
    """Find the record type column from a sequence of column names"""
    record_type_columns = [
        t
        for t in columnnames
        if t.lower().replace("_", "") in ("recordtype", "recordtypeid")
    ]
    if len(record_type_columns) > 1:
        raise DataGenError(
            f"Multiple record type columns for {tablename}: {record_type_columns}"
        )

    if len(record_type_columns) == 1:
        return record_type_columns[0]



import contextlib
from typing import Callable, ContextManager, Iterator, TypeVar

T = TypeVar("T")

# until https://github.com/agronholm/typeguard/issues/115 is fully fixed
# or typeguard is removed, we need this hack.

# The fix is on typeguard/master and should be merged in Typeguard 3.0
#
# We need this feature to be moved from UNRELEASED to RELEASED:
#
#    Changed the import hook to append @typechecked to the
#    decorator list instead of inserting it as the first decorator
#   (fixes type checking inconsistencies with mypy regarding at least
#    @contextmanager, probably others too)


def contextmanager(
    func: Callable[..., Iterator[T]]
) -> Callable[..., ContextManager[T]]:
    result = contextlib.contextmanager(func)
    result.__annotations__ = {**func.__annotations__, "return": ContextManager[T]}
    return result



from abc import abstractmethod, ABC
import json
from tempfile import TemporaryDirectory
import csv
import subprocess
import datetime
import sys
from decimal import Decimal
from pathlib import Path
from collections import namedtuple, defaultdict
from typing import Dict, TextIO, Union, Optional, Mapping, Callable, Sequence, cast
import typing as T
from warnings import warn

from sqlalchemy import (
    create_engine,
    MetaData,
    Column,
    Integer,
    Table,
    Unicode,
    func,
    inspect,
)
from sqlalchemy.ext.automap import automap_base
from sqlalchemy.orm import create_session
from sqlalchemy.engine import Engine
from sqlalchemy.sql import select

from .data_gen_exceptions import DataGenError

from .object_rows import ObjectRow, ObjectReference
from .parse_recipe_yaml import TableInfo


def noop(x):
    return x


def format_datetime(dt: datetime.datetime):
    """Format into the Salesforce-preferred syntax."""
    return dt.isoformat(timespec="seconds")


def simplifier_encoder(field_value):
    return field_value.simplify()


class OutputStream(ABC):
    """Common base class for all output streams"""

    count = 1
    flush_limit = 1000
    commit_limit = 10000
    encoders: Mapping[type, Callable] = {
        str: str,
        int: int,
        float: float,
        datetime.date: noop,
        datetime.datetime: noop,
        type(None): noop,
        bool: int,
        Decimal: str,
    }
    uses_folder = False
    uses_path = False
    is_text = False

    def __init__(self, filename, **kwargs):
        pass

    def create_or_validate_tables(self, tables: Dict[str, TableInfo]) -> None:
        pass

    def flatten(
        self,
        sourcetable: str,
        fieldname: str,
        source_row_dict: Dict,
        target_object_row: Union[ObjectRow, ObjectReference],
    ) -> Union[str, int]:
        return target_object_row.id

    def flush(self):
        pass  # pragma: no cover   -- subclasses override

    def commit(self):
        pass  # pragma: no cover   -- subclasses override

    def cleanup(self, field_name, field_value, sourcetable, row):
        if isinstance(field_value, (ObjectRow, ObjectReference)):
            return self.flatten(sourcetable, field_name, row, field_value)
        else:
            encoder = self.encoders.get(type(field_value))
            if not encoder and hasattr(field_value, "simplify"):
                encoder = simplifier_encoder
            if not encoder:
                raise TypeError(  # pragma: no cover
                    f"No encoder found for {type(field_value)} in {self.__class__.__name__} "
                    f"for {field_name}, {field_value} in {sourcetable}"
                )
            return encoder(field_value)

    def write_row(self, tablename: str, row_with_references: Dict) -> None:
        row_cleaned_up_and_flattened = {
            field_name: self.cleanup(
                field_name, field_value, tablename, row_with_references
            )
            for field_name, field_value in row_with_references.items()
        }
        self.write_single_row(tablename, row_cleaned_up_and_flattened)
        if self.count % self.flush_limit == 0:
            self.flush()

        if self.count % self.commit_limit == 0:
            self.commit()

        self.count += 1

    @abstractmethod
    def write_single_row(self, tablename: str, row: Dict) -> None:
        """Write a single row to the stream"""
        pass

    def close(self, **kwargs) -> Optional[Sequence[str]]:
        """Close any resources the stream opened.

        Do not close file handles which were passed in!

        Return a list of messages to print out.
        """
        raise NotImplementedError()

    def __enter__(self, *args):
        return self

    def __exit__(self, *args):
        self.close()


class SmartStream:
    """Common code for managing stream/file opening/closing

    Expects to be initialized with either a file-like object with a `write` method,
    or a path (str or pathlib.Path) that can be opened using `open()`
    """

    mode = "wt"

    def __init__(self, stream_or_path=None, **kwargs):
        if stream_or_path and hasattr(stream_or_path, "write"):
            self.owns_stream = False
            self.stream = stream_or_path
        elif stream_or_path:
            self.stream = open(stream_or_path, self.mode)
            self.owns_stream = True
        elif stream_or_path is None:
            self.owns_stream = False
            self.stream = sys.stdout
        else:
            raise AssertionError(f"stream_or_path is {stream_or_path}")

    def write(self, data):
        self.stream.write(data)

    def close(self, **kwargs):
        if self.owns_stream:
            self.stream.close()
            return [f"Generated {self.stream.name}"]


class FileOutputStream(OutputStream):
    """Base class for all file/stream-based OutputStreams"""

    def __init__(self, stream_or_path=None, **kwargs):
        self.smart_stream = SmartStream(stream_or_path)
        self.write = self.smart_stream.write
        self.stream = self.smart_stream.stream

    def close(self, **kwargs):
        return self.smart_stream.close()


class SimpleFileOutputStream(FileOutputStream):
    """Debug-like output, but with raw data fields"""

    is_text = True

    def write_single_row(self, tablename: str, row: Dict) -> None:
        values = ", ".join([f"{key}={value}" for key, value in row.items()])
        self.write(f"{tablename}({values})\n")

    def flatten(
        self,
        sourcetable: str,
        fieldname: str,
        source_row_dict: Dict,
        target_object_row: Union[ObjectRow, ObjectReference],
    ) -> Union[str, int]:
        return f"{target_object_row._tablename}({target_object_row.id})"


class DebugOutputStream(SimpleFileOutputStream):
    """Simplified output for debugging Snowfakery files.

    Datetimes are in Salesforce format."""

    encoders: Mapping[type, Callable] = {
        **SimpleFileOutputStream.encoders,
        datetime.datetime: format_datetime,  # format into Salesforce-friendly syntax
    }


CSVContext = namedtuple("CSVContext", ["dictwriter", "file"])


class CSVOutputStream(OutputStream):
    """Output stream that generates a directory of CSV files."""

    encoders: Mapping[type, Callable] = {
        **OutputStream.encoders,
        datetime.datetime: format_datetime,  # format into Salesforce-friendly syntax
    }

    uses_folder = True

    def __init__(self, output_folder, **kwargs):
        super().__init__(None, **kwargs)
        self.target_path = Path(output_folder)
        if not Path.exists(self.target_path):
            Path.mkdir(self.target_path, exist_ok=True)

    def open_writer(self, table_name, table):
        file = open(self.target_path / f"{table_name}.csv", "w", newline="")
        writer = csv.DictWriter(file, list(table.fields.keys()) + ["id"])
        writer.writeheader()
        return CSVContext(dictwriter=writer, file=file)

    def create_or_validate_tables(self, tables: Dict[str, TableInfo]) -> None:
        self.writers = {
            table_name: self.open_writer(table_name, table)
            for table_name, table in tables.items()
        }

    def write_single_row(self, tablename: str, row: Dict) -> None:
        self.writers[tablename].dictwriter.writerow(row)

    def close(self, **kwargs) -> Optional[Sequence[str]]:
        messages = []
        for context in self.writers.values():
            context.file.close()
            messages.append(f"Created {context.file.name}")

        table_metadata = [
            {"url": f"{table_name}.csv"} for table_name, writer in self.writers.items()
        ]
        csv_metadata = {
            "@context": "http://www.w3.org/ns/csvw",
            "tables": table_metadata,
        }
        csvw_filename = self.target_path / "csvw_metadata.json"
        with open(csvw_filename, "w") as f:
            json.dump(csv_metadata, f, indent=2)
        messages.append(f"Created {csvw_filename}")
        return messages


class JSONOutputStream(FileOutputStream):
    encoders: Mapping[type, Callable] = {
        **OutputStream.encoders,
        datetime.date: str,
        datetime.datetime: str,
        bool: bool,
    }
    is_text = True

    def __init__(self, file, **kwargs):
        assert file
        super().__init__(file, **kwargs)
        self.first_row = True

    def write_single_row(self, tablename: str, row: Dict) -> None:
        if self.first_row:
            self.write("[")
            self.first_row = False
        else:
            self.write(",\n")
        values = {"_table": tablename, **row}
        self.write(json.dumps(values))

    def close(self, **kwargs) -> Optional[Sequence[str]]:
        if not self.first_row:
            self.write("]\n")
        return super().close()


class SqlDbOutputStream(OutputStream):
    """Output stream for talking to SQL Databases"""

    encoders: Mapping[type, Callable] = {
        **OutputStream.encoders,
        datetime.datetime: format_datetime,  # format into Salesforce-friendly syntax
    }

    should_close_session = False

    def __init__(self, engine: Engine, mappings: None = None, **kwargs):
        if mappings:  # pragma: no cover  -- should not be triggered.
            warn("Please do not pass mappings argument to __init__", DeprecationWarning)
        self.buffered_rows = defaultdict(list)
        self.table_info = {}
        self.engine = engine
        self.session = create_session(bind=self.engine, autocommit=False)
        self.metadata = MetaData()
        self.base = automap_base(metadata=self.metadata)

    @classmethod
    def from_url(cls, db_url: str, mappings: None = None):
        if mappings:  # pragma: no cover  -- should not be triggered.
            warn("Please do not pass mappings argument to from_url", DeprecationWarning)
        try:
            engine = create_engine(db_url)
        except ModuleNotFoundError as e:
            raise DataGenError(f"Cannot find a driver for your database: {e}")
        except Exception as e:
            raise DataGenError(f"Cannot connect to database: {e}")
        self = cls(engine)
        return self

    def write_single_row(self, tablename: str, row: Dict) -> None:
        # cache the value for later insert
        self.buffered_rows[tablename].append(row)

    def flush(self):
        with self.session.begin():
            self._flush_rows()
            self.session.flush()

    def _flush_rows(self):
        for tablename, (insert_statement, fallback_dict) in self.table_info.items():
            # Make sure every row has the same records per SQLAlchemy's rules

            # According to the SQL Alchemy docs, every dictionary in a set must
            # have the same keys.

            # This means that the INSERT statement will be more bloated but it
            # seems much more efficient than line-by-line inserts.
            values = [
                {
                    key: row[key] if key in row else fallback_dict[key]
                    for key in fallback_dict.keys()
                }
                for row in self.buffered_rows[tablename]
            ]
            if values:
                self.session.execute(insert_statement, values)
            self.buffered_rows[tablename] = []

    def commit(self):
        if any(self.buffered_rows):
            self.flush()

    def close(self, **kwargs) -> Optional[Sequence[str]]:
        self.commit()
        self.session.close()
        self.engine.dispose()

    def create_or_validate_tables(self, inferred_tables: Dict[str, TableInfo]) -> None:
        try:
            create_tables_from_inferred_fields(
                inferred_tables, self.engine, self.metadata
            )
        except Exception as e:
            raise DataGenError(f"Cannot write to database: {e}")
        self.metadata.create_all(bind=self.engine)
        self.base.prepare(autoload_with=self.engine, reflect=True)

        # Setup table info used by the write-buffering infrastructure
        TableTuple = namedtuple("TableTuple", ["insert_statement", "fallback_dict"])

        for tablename, model in self.metadata.tables.items():
            if tablename in inferred_tables:
                table_info = TableTuple(
                    insert_statement=model.insert().inline(),
                    fallback_dict={
                        key: None for key in inferred_tables[tablename].fields.keys()
                    },
                )
                # id is special
                table_info.fallback_dict.setdefault("id", None)

                # See create_tables_from_inferred_fields to see what _sf_update_key are for
                if inferred_tables[tablename].has_update_keys:
                    table_info.fallback_dict.setdefault("_sf_update_key", None)
                self.table_info[tablename] = table_info


# backwards-compatible name for CCI
SqlOutputStream = SqlDbOutputStream


class SqlTextOutputStream(FileOutputStream):
    """Output stream to generate a SQL text file"""

    mode = "wt"
    is_text = True

    def __init__(self, stream_or_path=None, **kwargs):
        self.text_output = SmartStream(stream_or_path)
        self.tempdir = TemporaryDirectory()
        self.sql_db = self._init_db()
        super().__init__(**kwargs)

    def _init_db(self):
        "Initialize a db through an owned output stream"
        db_url = f"sqlite:///{self.tempdir.name}/tempdb.db"
        engine = create_engine(db_url)
        return SqlDbOutputStream(engine)

    def write_single_row(self, tablename: str, row: Dict) -> None:
        self.sql_db.write_single_row(tablename, row)
        # it might seem logical to try and output the raw SQL here
        # but it gets messy due to limitations of SQL Alchemy
        # https://docs.sqlalchemy.org/en/13/faq/sqlexpressions.html#rendering-bound-parameters-inline
        # in particular datetime values do not render properly without extra code
        # perhaps there are other, similar, undiscovered limitations.

    def create_or_validate_tables(self, tables: Dict[str, TableInfo]) -> None:
        self.sql_db.create_or_validate_tables(tables)

    def flush(self):
        self.sql_db.flush()

    def _dump_db(self):
        "Dump a database as a sql file"
        self.sql_db.commit()
        con = self.sql_db.engine.raw_connection()

        # https://docs.python.org/3/library/sqlite3.html#sqlite3.Connection.iterdump
        for line in con.iterdump():  # type: ignore
            assert self.text_output.stream
            self.text_output.stream.write("%s\n" % line)

        con.close()

    def close(self, *args, **kwargs):
        self._dump_db()
        self.sql_db.close(*args, **kwargs)
        self.text_output.close(*args, **kwargs)
        self.tempdir.cleanup()


def create_tables_from_inferred_fields(
    tables: T.Dict[str, TableInfo], engine, metadata
):
    """Create tables based on dictionary of tables->field-list."""
    with engine.connect() as conn:
        inspector = inspect(engine)
        for table_name, table in tables.items():
            columns = [Column(field_name, Unicode(255)) for field_name in table.fields]
            id_column_as_list = [
                column for column in columns if column.name.lower() == "id"
            ]

            # this code primarily supports using this function for
            # the SOQLDataSet and other datasets where ID may already exists
            if id_column_as_list:
                id_column = id_column_as_list[0]
            else:
                id_column = Column(
                    "id", Integer(), primary_key=True, autoincrement=True
                )

            # add a column keeping track of what update_key was specified by
            # the template. This allows multiple templates to have different
            # update_keys.
            if table.has_update_keys:
                columns.append(Column("_sf_update_key", Unicode(255)))

            t = Table(table_name, metadata, id_column, *columns)

            if inspector.has_table(table_name):
                stmt = select(func.count(t.c.id))
                count = conn.execute(stmt).first()[0]
                if count > 0:
                    raise DataGenError(
                        f"Table already exists and has data: {table_name} in {engine.url}",
                    )


def find_name_in_dict(d):
    "Try to find a key that is semantically a 'name' for diagramming purposes."
    keys = {k.lower().replace("_", ""): k for k in d.keys()}
    if "name" in keys:
        return d[keys["name"]]
    elif "firstname" in keys or "lastname" in keys:
        firstname = d[keys.get("firstname")] if keys.get("firstname") else ""
        lastname = d[keys.get("lastname")] if keys.get("lastname") else ""
        return " ".join([firstname, lastname])
    elif "name" in str(" ".join(d.keys())):
        namekey = [k for k in d.keys() if "name" in k][0]
        return d[namekey]
    elif "id" in keys:
        return d[keys["id"]]


class GraphvizOutputStream(FileOutputStream):
    """Generates a Graphviz .dot file"""

    is_text = True

    def __init__(self, path, **kwargs):
        import gvgen

        super().__init__(path, **kwargs)

        self.nodes = {}
        self.links = []
        self.G = gvgen.GvGen()
        self.G.styleDefaultAppend("fontsize", "10")
        self.G.styleDefaultAppend("style", "filled")
        self.G.styleDefaultAppend("fillcolor", "#009EDB")
        self.G.styleDefaultAppend("fontcolor", "#FFFFFF")
        self.G.styleDefaultAppend("height", "0.75")
        self.G.styleDefaultAppend("width", "0.75")
        self.G.styleDefaultAppend("widshapeth", "circle")

    def flatten(
        self,
        sourcetable: str,
        fieldname: str,
        source_row_dict: Dict,
        target_object_row: Union[ObjectRow, ObjectReference],
    ) -> Union[str, int]:
        source = (sourcetable, source_row_dict["id"])
        target = (target_object_row._tablename, target_object_row.id)
        self.links.append((fieldname, source, target))
        return ""

    def generate_node_name(
        self, tablename: str, rowname, id: Optional[int] = None
    ) -> str:
        rowname = (", " + rowname) if rowname and rowname != id else ""
        return f"{tablename}({id}{rowname})"

    def write_single_row(self, tablename: str, row: Dict) -> None:
        node_name = self.generate_node_name(
            tablename, find_name_in_dict(row), row["id"]
        )
        self.nodes[tablename, row["id"]] = self.G.newItem(node_name)

    def close(self, **kwargs) -> Optional[Sequence[str]]:
        for fieldname, source, target in self.links:
            mylink = self.G.newLink(self.nodes[source], self.nodes[target])
            self.G.propertyAppend(mylink, "label", fieldname)
        self.G.dot(cast(TextIO, self.stream))
        return super().close()


DOT_MISSING_MESSAGE = """
Could not find `dot` executable.

Please install graphviz and ensure that the command `dot` is available.
For example, on Mac you could try `brew install graphviz`
On Windows you could try `winget install graphviz` or `choco install graphviz`
Other installation options are here: http://www.graphviz.org/download/

If you have installed graphviz but Snowfakery cannot find it, perhaps you
will need to use Snowfakery to generate a dotfile (--output-file=out.dot)
and then you can convert it to another format yourself as described here:
https://stackoverflow.com/a/1494495/113477

If your data is not private, you could even use one of the online
converters that you can find by searching the Web for
"convert dot file to png online".
"""


class ImageOutputStream(OutputStream):
    """Output an Image file in a graphviz supported format."""

    mode = "wb"
    uses_path = True

    def __init__(self, path, format, **kwargs):
        self.path = path
        self.format = format
        self.tempdir = TemporaryDirectory()
        self.dotfile = Path(self.tempdir.name) / "temp.dot"
        self.gv_os = GraphvizOutputStream(self.dotfile)

    def write_single_row(self, *args, **kwargs):
        return self.gv_os.write_single_row(*args, **kwargs)

    def flatten(self, *args, **kwargs):
        return self.gv_os.flatten(*args, **kwargs)

    def close(self, **kwargs) -> Optional[Sequence[str]]:
        self.gv_os.close()
        assert self.dotfile.exists()
        rc = self._render(self.dotfile, self.path)
        self.tempdir.cleanup()
        return rc or [f"Generated {self.path}"]

    def _render(self, dotfile, outfile):
        assert dotfile.exists()
        try:
            out = subprocess.Popen(
                ["dot", "-T" + self.format, dotfile, "-o" + str(outfile)],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
            )
            stdout, stderr = out.communicate()
            stdout = [stdout.decode("ASCII")] if stdout else []
            stderr = [stderr.decode("ASCII")] if stderr else []
        except FileNotFoundError:
            return [DOT_MISSING_MESSAGE]
        return stdout + stderr


class MultiplexOutputStream(OutputStream):
    """Generate multiple output streams at once."""

    def __init__(self, outputstreams, **kwargs):
        self.outputstreams = outputstreams
        super().__init__(None, **kwargs)

    def create_or_validate_tables(self, tables: Dict[str, TableInfo]) -> None:
        for stream in self.outputstreams:
            stream.create_or_validate_tables(tables)

    def write_row(self, tablename: str, row_with_references: Dict) -> None:
        for stream in self.outputstreams:
            stream.write_row(tablename, row_with_references)

    def close(self, **kwargs) -> Optional[Sequence[str]]:
        for stream in self.outputstreams:
            stream.close()

    def write_single_row(self, tablename: str, row: Dict) -> None:
        raise NotImplementedError()  # should never be called



import random
import sys
from ast import literal_eval
from datetime import date, datetime
from functools import lru_cache
from datetime import timezone
from typing import Any, List, Tuple, Union

import dateutil.parser
from dateutil.relativedelta import relativedelta
from faker import Faker
from faker.providers.date_time import Provider as DateProvider

import snowfakery.data_generator_runtime  # noqa
from snowfakery.fakedata.fake_data_generator import UTCAsRelDelta, _normalize_timezone
from snowfakery.object_rows import ObjectReference
from snowfakery.plugins import PluginContext, SnowfakeryPlugin, lazy, memorable
from snowfakery.row_history import RandomReferenceContext
from snowfakery.standard_plugins.UniqueId import UniqueId
from snowfakery.utils.template_utils import StringGenerator

from .data_gen_exceptions import DataGenError

FieldDefinition = "snowfakery.data_generator_runtime_object_model.FieldDefinition"

# It might make more sense to use context vars for context handling when
# Python 3.6 is out of the support matrix.


def parse_weight_str(context: PluginContext, weight_value) -> float:
    """For constructs like:

    - choice:
        probability: 60%
        pick: Closed Won

    Render and convert the 60% to just 60.
    """
    weight_str = context.evaluate(weight_value)
    if isinstance(weight_str, str):
        weight_str = weight_str.rstrip("%")
    return float(weight_str)


def weighted_choice(choices: List[Tuple[float, object]]):
    """Selects from choices based on their weights"""
    weights = [weight for weight, value in choices]
    options = [value for weight, value in choices]
    return random.choices(options, weights, k=1)[0]


@lru_cache(maxsize=512)
def parse_date(d: Union[str, datetime, date]) -> date:
    if isinstance(d, datetime):
        return d.date()
    elif isinstance(d, date):
        return d

    return dateutil.parser.parse(d).date()


@lru_cache(maxsize=512)
def parse_datetimespec(d: Union[str, datetime, date]) -> datetime:
    """Parse a string, datetime or date into a datetime."""
    if isinstance(d, datetime):
        if not d.tzinfo:
            d = d.replace(tzinfo=timezone.utc)
        return d
    elif isinstance(d, str):
        if d == "now":
            return datetime.now(tz=timezone.utc)
        elif d == "today":
            return datetime.combine(
                date.today(), datetime.min.time(), tzinfo=timezone.utc
            )
        d = dateutil.parser.parse(d)
        if not d.tzinfo:
            d = d.replace(tzinfo=timezone.utc)
        return d
    elif isinstance(d, date):
        return datetime.combine(d, datetime.min.time(), tzinfo=timezone.utc)


def render_boolean(context: PluginContext, value: FieldDefinition) -> bool:
    val = context.evaluate(value)
    if isinstance(val, str):
        val = literal_eval(val)

    return bool(val)


class StandardFuncs(SnowfakeryPlugin):
    class Functions:
        int = int
        # use ONLY for random_dates
        # anything else should use the Faker from the Interpreter
        # which is locale-scoped.
        _faker_for_dates = Faker(use_weighting=False)
        _uidgen = None

        def __init__(self, *args, **kwargs):
            self.snowfakery_filename = StringGenerator(self._snowfakery_filename)
            self.unique_id = StringGenerator(self._unique_id)
            self.unique_alpha_code = StringGenerator(self._unique_alpha_code)

        def date(
            self,
            datespec=None,
            *,
            year: Union[str, int] = None,
            month: Union[str, int] = None,
            day: Union[str, int] = None,
        ):
            """A YAML-embeddable function to construct a date from strings or integers"""
            if datespec:
                if any((day, month, year)):
                    raise DataGenError(
                        "Should not specify a date specification and also day or month or year."
                    )
                return parse_date(datespec)
            else:
                return date(year, month, day)

        def datetime(
            self,
            datetimespec=None,
            *,
            year: Union[str, int] = None,
            month: Union[str, int] = None,
            day: Union[str, int] = None,
            hour=0,
            minute=0,
            second=0,
            microsecond=0,
            timezone=UTCAsRelDelta,
        ):
            """A YAML-embeddable function to construct a datetime from strings or integers"""
            timezone = _normalize_timezone(timezone)
            if datetimespec:
                if any((day, month, year, hour, minute, second, microsecond)):
                    raise DataGenError(
                        "Should not specify a date specification and also other parameters."
                    )
                dt = parse_datetimespec(datetimespec)
                dt = dt.replace(tzinfo=timezone)
            elif not (any((year, month, day, hour, minute, second, microsecond))):
                # no dt specification provided at all...just use now()
                dt = datetime.now(timezone)
            else:
                dt = datetime(
                    year, month, day, hour, minute, second, microsecond, timezone
                )

            return dt

        def date_between(self, *, start_date, end_date, timezone=UTCAsRelDelta):
            """A YAML-embeddable function to pick a date between two ranges"""

            def try_parse_date(d):
                if not isinstance(d, str) or not DateProvider.regex.fullmatch(d):
                    try:
                        d = parse_date(d)
                    except Exception:  # let's hope its something faker can parse
                        pass
                return d

            start_date = try_parse_date(start_date)
            end_date = try_parse_date(end_date)

            try:
                return self._faker_for_dates.date_between(start_date, end_date)
            except ValueError as e:
                if "empty range" not in str(e):
                    raise
            # swallow empty range errors per Python conventions

        def datetime_between(self, *, start_date, end_date, timezone=UTCAsRelDelta):
            """A YAML-embeddable function to pick a datetime between two ranges"""

            start_date = self.datetime(start_date)
            end_date = self.datetime(end_date)

            timezone = _normalize_timezone(timezone)
            if end_date < start_date:
                raise DataGenError("End date is before start date")

            return self._faker_for_dates.date_time_between(
                start_date, end_date, tzinfo=timezone
            )

        def i18n_fake(self, locale: str, fake: str):
            # deprecated by still here for backwards compatibility
            faker = Faker(locale, use_weighting=False)
            func = getattr(faker, fake)
            return func()

        def random_number(self, min: int, max: int, step: int = 1) -> int:
            """Pick a random number between min and max like Python's randint."""
            return random.randrange(min, max + 1, step)

        def reference(
            self, x: Any = None, object: str = None, id: Union[str, int] = None
        ):
            """YAML-embeddable function to Reference another object."""
            if x is not None:
                return self._reference_from_scalar(x)
            elif object:
                try:
                    id = int(id)
                except (ValueError, TypeError):
                    raise DataGenError("Cannot interpret `id` as an integer")
                return ObjectReference(object, id)

        def _reference_from_scalar(self, x: Any):
            if hasattr(x, "id"):  # reference to an object with an id
                target = x
            elif isinstance(x, str):  # name of an object
                # allows dotted paths
                parts = x.split(".")
                target = self.context.field_vars().get(parts.pop(0))

                for part in parts:
                    try:
                        target = getattr(target, part)
                    except AttributeError:
                        raise DataGenError(
                            f"Expression cannot be evaluated `{x}``. Problem before `{part}`:\n {target}",
                            None,
                            None,
                        )

                if not target:
                    raise DataGenError(f"Cannot find an object named {x}", None, None)
                if not getattr(target, "id", None):
                    raise DataGenError(
                        f"Reference to incorrect object type {target}", None, None
                    )
            else:
                raise DataGenError(
                    f"Can't get reference to object of type {type(x)}: {x}"
                )

            return target

        @lazy
        def random_choice(self, *choices, **kwchoices):
            """Template helper for random choices.

            Supports structures like this:

            random_choice:
                - a
                - b
                - ${{c}}

            Or like this:

            random_choice:
                - choice:
                    pick: A
                    probability: 50%
                - choice:
                    pick: A
                    probability: 50%

            Probabilities are really just weights and don't need to
            add up to 100.

            Pick-items can have arbitrary internal complexity.

            Pick-items are lazily evaluated.
            """

            # very occasionally single-item choices are useful
            use_choices = len(choices) >= 1

            # very occasionally single-item choices are useful
            use_kwchoices = len(kwchoices) >= 1

            if not (use_choices or use_kwchoices):
                raise ValueError("No choices supplied!")
            elif use_choices and use_kwchoices:
                raise ValueError("Both choices and probabilities supplied!")
            elif use_choices:
                if getattr(choices[0], "function_name", None) == "choice":
                    choices = [self.context.evaluate_raw(choice) for choice in choices]
                    rc = weighted_choice(choices)
                else:
                    rc = random.choice(choices)
                if hasattr(rc, "render"):
                    rc = self.context.evaluate_raw(rc)
            else:
                assert use_kwchoices and not use_choices
                choices = [
                    (parse_weight_str(self.context, value), key)
                    for key, value in kwchoices.items()
                ]
                rc = weighted_choice(choices)

            return rc

        @lazy
        def choice(
            self,
            pick,
            probability: FieldDefinition = None,
            when: FieldDefinition = None,
        ):
            """Supports the choice: sub-items used in `random_choice` or `if`"""
            if probability:
                probability = parse_weight_str(self.context, probability)
            return probability or when, pick

        @memorable
        def random_reference(
            self,
            to: str,
            *,
            parent: str = None,
            scope: str = "current-iteration",
            unique: bool = False,
        ) -> "RandomReferenceContext":
            """Select a random, already-created row from 'sobject'

            - object: Owner
              count: 10
              fields:
                name: fake.name
            - object: Pet
              count: 10
              fields:
                ownedBy:
                    random_reference:
                        Owner

            See the docs for more info.
            """
            return RandomReferenceContext(
                self.context.interpreter.row_history, to, scope, unique
            )

        @lazy
        def if_(self, *choices: FieldDefinition):
            """Template helper for conditional choices.

            Supports structures like this:

            if:
                - choice:
                    when: ${{something}}
                    pick: A
                - choice:
                    when: ${{something}}
                    pick: B

            Pick-items can have arbitrary internal complexity.

            Pick-items are lazily evaluated.
            """
            if not choices:
                raise ValueError("No choices supplied!")

            choices = [self.context.evaluate_raw(choice) for choice in choices]
            for when, choice in choices[:-1]:
                if when is None:
                    raise SyntaxError(
                        "Every choice except the last one should have a when-clause"
                    )
            true_choices = (
                choice
                for when, choice in choices
                if when and render_boolean(self.context, when)
            )
            rc = next(true_choices, choices[-1][-1])  # default to last choice
            if hasattr(rc, "render"):
                rc = self.context.evaluate_raw(rc)
            return rc

        def _snowfakery_filename(self):
            template = self.context.field_vars()["template"]
            return template.filename

        @property
        def _unique_id_generator(self):
            if not self._uidgen:
                self._uidgen = UniqueId(self.context.interpreter).custom_functions()

            return self._uidgen

        def _unique_id(self):
            return self._unique_id_generator.default_uniqifier.unique_id

        def _unique_alpha_code(self):
            return self._unique_id_generator.default_alpha_code_generator.unique_id

        def debug(self, value):
            msg = f"DEBUG - {value} ({type(value)})\n"
            sys.stderr.write(msg)
            return value

    setattr(Functions, "if", Functions.if_)
    setattr(Functions, "relativedelta", relativedelta)
    setattr(Functions, "NULL", None)
    setattr(Functions, "null", None)
    setattr(Functions, "Null", None)



from enum import Enum, auto
import typing as T

import yaml
import snowfakery  # noqa
from .utils.yaml_utils import SnowfakeryDumper
from contextvars import ContextVar

if T.TYPE_CHECKING:
    from snowfakery.data_generator_runtime import IdManager
else:
    IdManager = "snowfakery.data_generator_runtime.IdManager"
RowHistoryCV = ContextVar("RowHistory")


class ObjectRow:
    """Represents a single row

    Uses __getattr__ so that the template evaluator can use dot-notation."""

    yaml_loader = yaml.SafeLoader
    yaml_dumper = SnowfakeryDumper
    yaml_tag = "!snowfakery_objectrow"

    # be careful changing these slots because these objects must be serializable
    # to YAML and JSON
    __slots__ = ["_tablename", "_values", "_child_index"]

    def __init__(self, tablename, values=(), index=0):
        self._tablename = tablename
        self._values = values
        self._child_index = index

    def __getattr__(self, name):
        try:
            return self._values[name]
        except KeyError:
            raise AttributeError(name)

    # prefer this to .id or _values["id"] so we can
    # one day move away from having "id" in the user's
    # namespace
    @property
    def _id(self):
        return self._values["id"]

    def __str__(self):
        return str(self.id)

    def __repr__(self):
        try:
            return f"<ObjectRow {self._tablename} {self.id}>"
        except Exception:
            return super().__repr__()

    def __getstate__(self):
        """Get the state of this ObjectRow for serialization.

        Do not include related ObjectRows because circular
        references in serialization formats cause problems."""
        values = {k: v for k, v in self._values.items() if not isinstance(v, ObjectRow)}
        return {"_tablename": self._tablename, "_values": values}

    def __setstate__(self, state):
        for slot, value in state.items():
            setattr(self, slot, value)


class ObjectReference(yaml.YAMLObject):
    def __init__(self, tablename: str, id: int):
        self._tablename = tablename
        self.id = id


class LazyLoadedObjectReference(ObjectReference):
    _data = None

    def __init__(
        self,
        tablename: str,
        id: int,
        sql_tablename: str,
    ):
        self._tablename = tablename
        self.sql_tablename = sql_tablename
        self.id = id

    def __getattr__(self, attrname):
        if attrname.endswith("__"):  # pragma: no cover
            raise AttributeError(attrname)
        if self._data is None:
            row_history = RowHistoryCV.get()
            self._data = row_history.load_row(self.sql_tablename, self.id)
        return self._data[attrname]


class SlotState(Enum):
    """The current state of a NicknameSlot.

    UNUSED=empty, ALLOCATED=referenced, CONSUMED=object generated"""

    UNUSED = auto()
    ALLOCATED = auto()
    CONSUMED = auto()


class NicknameSlot(ObjectReference):
    """A slot that represents a Nickname or Tablename"""

    _tablename: str
    id_manager: IdManager
    allocated_id: T.Union[T.Optional[int], SlotState] = None

    def __init__(self, tablename: str, id_manager: IdManager):
        self._tablename = tablename
        self.id_manager = id_manager

    @property
    def id(self):
        "Get an id corresponding to this slot. Generate one if necessary."
        if self.allocated_id is None:
            self.allocated_id = self.id_manager.generate_id(self._tablename)
        return self.allocated_id

    def consume_slot(self):
        "Mark a slot as filled by an object and return its id for use by the object."
        rc = self.allocated_id
        self.allocated_id = SlotState.CONSUMED
        return rc

    @property
    def status(self):
        "Is the slot empty/unreferenced, allocated/referenced or consumed/used"
        if self.allocated_id is None:
            return SlotState.UNUSED
        elif isinstance(self.allocated_id, int):
            return SlotState.ALLOCATED
        elif self.allocated_id == SlotState.CONSUMED:
            return SlotState.CONSUMED

    def __repr__(self):
        return f"<NicknameSlot {self._tablename} {self.status} {self.allocated_id}>"



from base64 import b64encode
from logging import getLogger
from pathlib import Path
from random import randrange
from tempfile import TemporaryDirectory

from snowfakery import data_gen_exceptions as exc
from snowfakery.data_gen_exceptions import (
    DataGenError,
    DataGenNameError,
    DataGenValueError,
)
from snowfakery.data_generator_runtime_object_model import (
    FieldFactory,
    ObjectTemplate,
    SimpleValue,
    StructuredValue,
)
from snowfakery.output_streams import SqlDbOutputStream
from snowfakery.parse_recipe_yaml import TableInfo
from snowfakery.plugins import (
    ParserMacroPlugin,
    PluginOption,
    PluginResult,
    SnowfakeryPlugin,
    memorable,
)
from snowfakery.standard_plugins.datasets import (
    DatasetBase,
    DatasetPluginBase,
    sql_dataset,
)

MAX_SALESFORCE_OFFSET = 2000  # Any way around this?

# the option name that the user specifies on the CLI or API is just "org_name"
# but using this long name internally prevents us from clashing with the
# user's variable names.
plugin_option_org_name = (
    "snowfakery.standard_plugins.Salesforce.SalesforceQuery.org_name"
)
plugin_option_org_config = (
    "snowfakery.standard_plugins.Salesforce.SalesforceQuery.org_config"
)
plugin_option_project_config = (
    "snowfakery.standard_plugins.Salesforce.SalesforceQuery.project_config"
)


class SalesforceConnection:
    """Helper layer above simple_salesforce and salesforce_bulk"""

    _sf = None

    def __init__(self, get_project_config_and_org_config):
        self.get_project_config_and_org_config = get_project_config_and_org_config
        self.logger = getLogger(__name__)

    @property
    def sf(self):
        """simple_salesforce client"""
        if not self._sf:
            project_config, org_config = self.get_project_config_and_org_config()
            self._sf, self._bulk = self._get_sf_clients(project_config, org_config)
        return self._sf

    @property
    def bulk(self):
        """salesforce_bulk client"""
        self.sf  # initializes self._bulk as a side-effect
        return self._bulk

    def query(self, *args, **kwargs):
        """Query Salesforce through simple_salesforce"""
        return self.sf.query(*args, **kwargs)

    def query_single_record(self, query):
        """Query Salesforce through simple_salesforce and
        validate that query returns 1 and only 1 record"""
        qr = self.sf.query(query)
        records = qr.get("records")
        if not records:
            raise DataGenValueError(f"No records returned by query {query}", None, None)
        elif len(records) > 1:  # pragma: no cover
            raise DataGenValueError(
                f"Multiple records returned by query {query}", None, None
            )
        record = records[0]
        if "attributes" in record:
            del record["attributes"]
        if len(record.keys()) == 1:
            return tuple(record.values())[0]
        else:
            return PluginResult(record)

    def compose_query(self, context_name, **kwargs):
        kwargs = kwargs.copy()
        fields = kwargs.pop("fields", None)
        sobject = kwargs.pop("from", None)
        where = kwargs.pop("where", None)

        if not fields:
            raise DataGenError(f"{context_name} needs a 'fields' list")

        if not sobject:
            raise DataGenError(f"{context_name} needs a 'from'")

        if kwargs:
            raise DataGenError(
                f"Unknown argument in {context_name}: {tuple(kwargs.keys())}"
            )

        query = f"SELECT {fields} FROM {sobject} "
        if where:
            query += f" WHERE {where}"
        return query

    @staticmethod
    def _get_sf_clients(project_config, org_config):
        from cumulusci.salesforce_api.utils import get_simple_salesforce_connection

        sf = get_simple_salesforce_connection(project_config, org_config)
        return sf, _init_bulk(sf, org_config)


def _init_bulk(sf, org_config):
    from salesforce_bulk import SalesforceBulk

    return SalesforceBulk(
        host=org_config.instance_url.replace("https://", "").rstrip("/"),
        sessionId=org_config.access_token,
        API_version=sf.sf_version,
    )


def check_orgconfig(config):
    from cumulusci.core.config import BaseConfig

    if isinstance(config, BaseConfig):
        return config
    raise TypeError(f"Should be a CCI Config, not {type(config)}")


class SalesforceConnectionMixin:
    _sf_connection = None
    _runtime = None
    allowed_options = [
        PluginOption(plugin_option_org_name, str),
        PluginOption(plugin_option_org_config, check_orgconfig),
        PluginOption(plugin_option_project_config, check_orgconfig),
    ]

    @property
    def sf_connection(self):
        assert self.context
        if not self._sf_connection:
            self._sf_connection = SalesforceConnection(
                self.get_project_config_and_org_config
            )
        return self._sf_connection

    def get_project_config_and_org_config(self):
        fieldvars = self.context.field_vars()
        project_config = fieldvars.get(plugin_option_project_config)
        org_config = fieldvars.get(plugin_option_org_config)

        if not project_config or not org_config:
            project_config, org_config = self._get_org_info_from_cli_keychain()

        return project_config, org_config

    def _get_org_info_from_cli_keychain(self):
        org_name = self.get_org_name()  # from command line argument
        runtime = self._get_CliRuntime()  # from CCI CliRuntime
        name, org_config = runtime.get_org(org_name)
        return runtime.project_config, org_config

    def _get_CliRuntime(self):
        if self._runtime:
            return self._runtime  # pragma: no cover

        try:
            from cumulusci.cli.runtime import CliRuntime

            self._runtime = CliRuntime(load_keychain=True)
            return self._runtime
        except Exception as e:  # pragma: no cover
            raise DataGenError("CumulusCI Runtime cannot be loaded", *e.args)

    def get_org_name(self):
        """Look up the org_name in the scope"""
        fieldvars = self.context.field_vars()
        try:
            return fieldvars[plugin_option_org_name]
        except KeyError:
            raise DataGenNameError(
                "Orgname is not specified. Use --plugin-option org_name <yourorgname>",
                None,
                None,
            )


class Salesforce(ParserMacroPlugin, SnowfakeryPlugin, SalesforceConnectionMixin):
    def __init__(self, *args, **kwargs):
        args = args or [None]
        super().__init__(*args, **kwargs)

    def SpecialObject(self, context, args) -> ObjectTemplate:
        """Currently there is only one special object defined: PersonContact"""
        sobj, nickname = self._parse_special_args(args)
        line_info = context.line_num()
        if sobj == "PersonContact":
            return self._render_person_contact(context, sobj, nickname, line_info)
        else:
            raise exc.DataGenError(
                f"Unknown special object '{sobj}'. Did you mean 'PersonContact'?",
                None,
                None,
            )

    def _render_person_contact(self, context, sobj, nickname, line_info):
        """Generate the code to render a person contact as CCI expects.

        Code generation is a better strategy for this than a runtime
        plugin because some analysis of the table structures happens
        at parse time.
        """
        fields = [
            FieldFactory(
                "IsPersonAccount",
                SimpleValue("true", **line_info),
                **line_info,
            ),
            FieldFactory(
                "AccountId",
                StructuredValue("reference", ["Account"], **line_info),
                **line_info,
            ),
        ]
        new_template = ObjectTemplate(
            sobj,
            filename=line_info["filename"],
            line_num=line_info["line_num"],
            nickname=nickname,
            fields=fields,
        )
        context.register_template(new_template)
        return new_template

    def _parse_special_args(self, args):
        """Parse args of SpecialObject"""
        nickname = None
        if isinstance(args, str):
            sobj = args
        elif isinstance(args, dict):
            sobj = args["name"]
            if not isinstance(sobj, str):
                raise exc.DataGenError(
                    f"`name` argument should be a string, not `{sobj}`: ({type(sobj)})"
                )
            nickname = args.get("nickname")
            if nickname and not isinstance(nickname, str):
                raise exc.DataGenError(
                    f"`nickname` argument should be a string, not `{nickname}``: ({type(sobj)})"
                )
        else:
            raise exc.DataGenError(
                f"`name` argument should be a string, not `{args}``: ({type(args)})"
            )

        return sobj, nickname

    class Functions:
        @memorable
        def ProfileId(self, name):
            query = f"select Id from Profile where Name='{name}'"
            return self.context.plugin.sf_connection.query_single_record(query)

        Profile = ProfileId

        def ContentFile(self, file: str):
            template_path = Path(self.context.current_filename).parent

            with open(template_path / file, "rb") as data:
                return b64encode(data.read()).decode("ascii")


class SOQLDatasetImpl(DatasetBase):
    iterator = None
    tempdir = None

    def __init__(self, plugin, *args, **kwargs):
        from cumulusci.tasks.bulkdata.step import (
            DataOperationStatus,
            get_query_operation,
        )

        self.get_query_operation = get_query_operation
        self.DataOperationStatus = DataOperationStatus
        self.plugin = plugin
        super().__init__(*args, **kwargs)

    @property
    def sf_connection(self):
        return self.plugin.sf_connection

    def _load_dataset(self, iteration_mode, rootpath, kwargs):
        from cumulusci.tasks.bulkdata.step import DataApi

        query = self.sf_connection.compose_query("SOQLDataset", **kwargs)
        fields = kwargs.get("fields")
        sobject = kwargs.get("from")
        fieldnames = [f.strip() for f in fields.split(",")]
        qs = self.get_query_operation(
            sobject=sobject,
            fields=fieldnames,
            api_options={},
            context=self.sf_connection,
            query=query,
            api=DataApi.SMART,
        )

        try:
            qs.query()
        except Exception as e:
            raise DataGenError(f"Unable to query records for {query}: {e}") from e

        if qs.job_result.status is not self.DataOperationStatus.SUCCESS:
            raise DataGenError(
                f"Unable to query records for {query}: {','.join(qs.job_result.job_errors)}"
            )

        tempdir, iterator = create_tempfile_sql_db_iterator(
            iteration_mode, fieldnames, qs.get_results()
        )
        iterator.cleanup.push(tempdir)
        return iterator

    def close(self):
        pass


def create_tempfile_sql_db_iterator(mode, fieldnames, results):
    tempdir, db_url = _create_db(fieldnames, results)
    rc = sql_dataset(db_url, "data", mode)
    return tempdir, rc


def _create_db(fieldnames, results):
    tempdir = TemporaryDirectory()
    tempfile = Path(tempdir.name) / "queryresults.db"
    # TODO: try a real tempdb: "sqlite:///"
    dburl = f"sqlite:///{tempfile}"
    with SqlDbOutputStream.from_url(dburl) as db:
        ti = TableInfo("data")
        ti.fields = {fieldname: None for fieldname in fieldnames}
        db.create_or_validate_tables({"data": ti})
        for row in results:
            row_dict = {fieldname: result for fieldname, result in zip(fieldnames, row)}
            db.write_row("data", row_dict)
        db.flush()
        db.close()
    return tempdir, dburl


class SOQLDataset(SalesforceConnectionMixin, DatasetPluginBase):
    def __init__(self, *args, **kwargs):
        self.dataset_impl = SOQLDatasetImpl(self)
        super().__init__(*args, **kwargs)


class SalesforceQuery(SalesforceConnectionMixin, SnowfakeryPlugin):
    class Functions:
        @property
        def _sf_connection(self):
            return self.context.plugin.sf_connection

        def random_record(self, *args, fields="Id", where=None, **kwargs):
            """Query a random record."""
            context_vars = self.context.context_vars()
            context_vars.setdefault("count_query_cache", {})

            # "from" has to be handled separately because its a Python keyword
            query_from = self._parse_from_from_args(args, kwargs)

            # TODO: Test WHERE
            where_clause = f" WHERE {where}" if where else ""
            count_query = f"SELECT count() FROM {query_from}{where_clause}"
            count_result = self._sf_connection.query(count_query)

            count = count_result["totalSize"]
            mx = min(count, MAX_SALESFORCE_OFFSET)
            context_vars["count_query_cache"][count_query] = mx
            if mx < 1:
                raise DataGenError(
                    f"No records found matching {query_from}{where_clause}"
                )
            rand_offset = randrange(0, mx)
            query = f"SELECT {fields} FROM {query_from}{where_clause} LIMIT 1 OFFSET {rand_offset}"
            # todo: use CompositeParallelSalesforce to cache 200 at a time
            return self._sf_connection.query_single_record(query)

        @memorable
        def find_record(self, *args, fields="Id", where=None, **kwargs):
            """Find a particular record"""
            query_from = self._parse_from_from_args(args, kwargs)
            where_clause = f" WHERE {where}" if where else ""
            query = f"SELECT {fields} FROM {query_from}{where_clause} LIMIT 1"
            return self._sf_connection.query_single_record(query)

        def _parse_from_from_args(self, args, kwargs):
            query_from = None
            if kwargs:
                query_from = kwargs.pop("from", None)

                if kwargs:
                    raise ValueError(f"Unknown arguments: {tuple(kwargs.keys())}")
            elif args:
                if len(args) != 1 or not isinstance(args[0], str):
                    raise ValueError(f"Only one string argument allowed, not: {args}")
                query_from = args[0]

            if not query_from:
                raise ValueError("Must supply 'from:'")

            return query_from



from .plugins import (
    SnowfakeryPlugin,
    lazy,
    memorable,
    PluginResult,
    PluginResultIterator,
)
from .api import generate_data, SnowfakeryApplication

__all__ = (
    "generate_data",
    "SnowfakeryApplication",
    "SnowfakeryPlugin",
    "lazy",
    "memorable",
    "PluginResult",
    "PluginResultIterator",
)



from copy import copy


class DataGenError(Exception):
    prefix = (
        "An error occurred. If you would like to see a Python traceback, "
        "use the --debug-internals option."
    )

    def __init__(self, message, filename=None, line_num=None):
        self.message = message
        self.filename = filename
        self.line_num = line_num
        assert isinstance(filename, (str, type(None)))
        assert isinstance(line_num, (int, type(None)))
        super().__init__(self.message)

    def __str__(self):
        if self.line_num:
            location = f"\n near {self.filename}:{self.line_num}"
        elif self.filename:
            location = f"\n in {self.filename}"
        else:
            location = ""
        return f"{self.message}{location}"


class DataGenSyntaxError(DataGenError):
    pass


class DataGenYamlSyntaxError(DataGenSyntaxError):
    prefix = (
        "There is a problem with your YAML file.\n"
        + "Consider installing a YAML Validator Plugin for your editor.\n"
        + "For example, if you use VSCode, "
        + "https://marketplace.visualstudio.com/items?itemName=redhat.vscode-yaml \n"
        + "Or: https://marketplace.visualstudio.com/items?itemName=docsmsft.docs-yaml \n"
        + "The error is:\n"
    )
    pass


class DataGenNameError(DataGenError):
    pass


class DataGenValueError(DataGenError):
    pass


class DataGenImportError(DataGenError):
    pass


class DataGenTypeError(DataGenError):
    pass


def fix_exception(message, parentobj, e, args=(), kwargs=None):
    """Add filename and linenumber to an exception if needed"""
    filename, line_num = parentobj.filename, parentobj.line_num
    if isinstance(e, DataGenError):
        origmessage = e.message
    else:
        origmessage = str(e)

    message = message.format(*args, e=origmessage, **(kwargs or {}))
    if isinstance(e, DataGenError):
        e = copy(e)
        if not e.filename:
            e.filename = filename
        if not e.line_num:
            e.line_num = line_num
        e.message = message
        return e
    else:
        return DataGenError(message, filename, line_num)



import math
from snowfakery.plugins import SnowfakeryPlugin


class Math(SnowfakeryPlugin):
    def custom_functions(self, *args, **kwargs):
        "Expose math functions to Snowfakery"

        class MathNamespace:
            pass

        mathns = MathNamespace()
        mathns.__dict__ = math.__dict__.copy()

        mathns.pi = math.pi
        mathns.round = round
        mathns.min = min
        mathns.max = max

        return mathns



from itertools import count
from datetime import timedelta, date
import typing as T

from faker.providers.date_time import ParseError, Provider as DateProvider

from snowfakery import SnowfakeryPlugin
from snowfakery.plugins import PluginResultIterator, memorable
from snowfakery import data_gen_exceptions as exc


# TODO: merge this with template_funcs equivalent
def try_parse_date(d):
    from snowfakery.template_funcs import parse_date

    if not isinstance(d, str) or not DateProvider.regex.fullmatch(d):
        try:
            return parse_date(d)
        except Exception:  # let's hope its something faker can parse
            try:
                return DateProvider._parse_date(d)
            except ParseError as e:
                raise exc.DataGenValueError(d) from e


class NumberCounter(PluginResultIterator):
    def __init__(self, start=1, step=1):
        start = 1 if start is None else start
        self.counter = count(start, step)
        self.result = {
            "start": start,
            "step": step,
        }  # for serialization during continuation

    def next(self):
        return next(self.counter)


class DateCounter(PluginResultIterator):
    def __init__(self, start_date: T.Union[str, date] = "today", step: str = "+1d"):
        self.start_date = try_parse_date(start_date)
        if not self.start_date:  # pragma: no cover
            raise exc.DataGenError(f"Cannot parse start date {start_date}")
        step = DateProvider._parse_timedelta(step)
        self.counter = count(0, step)
        self.result = {
            "start": self.start_date,
            "step": step,
        }  # for serialization during continuation

    def next(self):
        offset = next(self.counter)
        return self.start_date + timedelta(seconds=offset)


class Counters(SnowfakeryPlugin):
    class Functions:
        @memorable
        def NumberCounter(
            self,
            _=None,
            *,
            start=1,
            step=1,
            name=None,
            parent=None,
        ):
            return NumberCounter(start, step)

        @memorable
        def DateCounter(
            self,
            _=None,
            *,
            start_date,
            step,
            name=None,
            parent=None,
        ):
            return DateCounter(start_date=start_date, step=step)



from os import fsdecode
from numbers import Number
from datetime import date
from contextlib import contextmanager
from pathlib import Path
from typing import IO, List, Dict, Union, Tuple, Any, Iterable, Sequence, Mapping
import typing as T
from warnings import warn

import yaml
from yaml.composer import Composer
from yaml.constructor import SafeConstructor
from yaml.error import YAMLError

from .data_generator_runtime_object_model import (
    ObjectTemplate,
    VariableDefinition,
    ForEachVariableDefinition,
    FieldFactory,
    SimpleValue,
    StructuredValue,
    Statement,
    Definition,
)
from snowfakery.standard_plugins.datasets import CSVDatasetLinearIterator
from snowfakery.plugins import resolve_plugins, LineTracker, ParserMacroPlugin
import snowfakery.data_gen_exceptions as exc
from snowfakery.utils.files import FileLike

SHARED_OBJECT = "#SHARED_OBJECT"

TemplateLike = Union[ObjectTemplate, StructuredValue]
###
#   The entry point to this is parse_recipe


class ParseResult:
    def __init__(
        self,
        options,
        tables: Mapping,
        statements: Sequence,
        plugins: Sequence = (),
        random_references: Sequence = (),
        version: int = None,
    ):
        self.options = options
        self.tables = tables
        self.statements = statements
        self.templates = [obj for obj in statements if isinstance(obj, ObjectTemplate)]
        self.plugins = plugins
        self.version = version
        self.random_references = random_references or []


class TableInfo:
    """Information we can infer about the table schema based on the templates.

    Note that a table can be referred to in more than one place, so this class
    unifies what we know about it.
    """

    def __init__(self, name):
        self.name = name
        self.fields = {}
        self.friends = {}
        self.has_update_keys = False
        self._templates = []

    def register(self, template: ObjectTemplate) -> None:
        self.fields.update(
            {
                field.name: field
                for field in template.fields
                if not field.name.startswith("__")
            }
        )
        self.friends.update(
            {
                friend.tablename: friend
                for friend in template.friends
                if hasattr(friend, "tablename")
            }
        )
        if template.update_key:
            self.has_update_keys = True
        self._templates.append(template)

    def __repr__(self) -> str:
        return f"<TableInfo name={self.name} fields={list(self.fields.keys())} friends={list(self.friends.keys())} templates={len(self._templates)}>"


class ParseContext:
    current_parent_object = None
    top_level = True

    def __init__(self):
        self.macros = {}
        self.plugins = []
        self.line_numbers = {}
        self.options = []
        self.table_infos = {}
        self.parser_macros_plugins = {}
        self.random_references = []

    def line_num(self, obj=None) -> Dict:
        if not obj:
            obj = self.current_parent_object
            assert obj

        # dicts should have __line__ keys
        try:
            return obj["__line__"]._asdict()
        except TypeError:
            pass

        # strings (and perhaps some other non-dict objects) should be tracked in
        # the line_numbers system
        try:
            my_line_num = self.line_numbers.get(id(obj))
            if my_line_num and my_line_num != SHARED_OBJECT:
                return my_line_num._asdict()
        except KeyError:  # pragma: no cover
            pass

        assert obj != self.current_parent_object  # check for no infinite loop
        return self.line_num(self.current_parent_object)

    @contextmanager
    def change_current_parent_object(self, obj: Dict):
        _old_parsed_template = self.current_parent_object
        self.current_parent_object = obj
        try:
            yield
        finally:
            self.current_parent_object = _old_parsed_template

    def register_template(self, template: ObjectTemplate) -> None:
        """Register templates for later use.

        We register templates so we can get a list of all fields that can
        be generated. This can be used to create a dynamic schema.
        """
        table_info = self.table_infos.get(template.tablename, None) or TableInfo(
            template.tablename
        )
        self.table_infos[template.tablename] = table_info
        table_info.register(template)

    @property
    def top_level(self):
        return self.current_parent_object is None


def removeline_numbers(dct: Dict) -> Dict:
    return {i: dct[i] for i in dct if i != "__line__"}


def _coerce_to_string(val, context):
    if isinstance(val, str):
        return val
    elif isinstance(val, (int, bool, date)):
        # a few functions accept keyword arguments that YAML interprets
        # as types other than string
        return str(val)
    else:
        if val is None:
            val = "null (None)"
        raise exc.DataGenSyntaxError(
            f"Cannot interpret key `{val}`` as string", **context.line_num()
        )


def parse_structured_value_args(
    args, context: ParseContext
) -> Union[List, Dict, SimpleValue, StructuredValue, ObjectTemplate]:
    """Structured values can be dicts or lists containing simple values or further structure."""
    if isinstance(args, dict):

        def as_str(name):
            return _coerce_to_string(name, context)

        with context.change_current_parent_object(args):
            return {
                as_str(name): parse_field_value(as_str(name), arg, context)
                for name, arg in args.items()
                if name != "__line__"
            }
    elif isinstance(args, list):
        return [
            parse_field_value(f"List element {i}", arg, context)
            for i, arg in enumerate(args)
        ]
    else:
        return parse_field_value("", args, context)


def parse_structured_value(name: str, field: Dict, context: ParseContext) -> Definition:
    """Parse something that might look like:

    {'choose': ['Option1', 'Option2', 'Option3', 'Option4'], '__line__': 9}

    or

    {'random_number': {'min': 10, 'max': 20, '__line__': 10} , '__line__': 9}
    """
    top_level = removeline_numbers(field).items()
    if not top_level:
        raise exc.DataGenSyntaxError(
            f"Strange datastructure ({field})", **context.line_num(field)
        )
    elif len(top_level) == 1:
        [[function_name, args]] = top_level
    elif len(top_level) > 1:
        top_level = list(top_level)
        function_name, blank = top_level.pop(0)
        args = dict(top_level)
        args["__line__"] = field.get("__line__")
    else:
        raise NotImplementedError("This should be unreachable!")

    plugin = None
    if "." in function_name:
        namespace, name = function_name.split(".")
        plugin = context.parser_macros_plugins.get(namespace)
    if plugin:
        try:
            func = getattr(plugin, name)
            rc = func(context, args)
            return rc
        except AttributeError:
            # we'll try to find it a different way below
            pass

    args = parse_structured_value_args(args, context)
    ret = StructuredValue(function_name, args, **context.line_num(field))
    if function_name == "random_reference":
        context.random_references.append(ret)  # need to track these for static analysis
    return ret


def parse_field_value(
    name: str,
    field,
    context: ParseContext,
) -> Union[SimpleValue, StructuredValue, ObjectTemplate]:
    if isinstance(field, (str, Number, date, type(None))):
        return SimpleValue(field, **(context.line_num(field) or context.line_num()))
    elif isinstance(field, dict) and field.get("object"):
        with context.change_current_parent_object(field):
            return parse_object_template(field, context)
    elif isinstance(field, dict):
        return parse_structured_value(name, field, context)
    elif isinstance(field, list) and len(field) == 1 and isinstance(field[0], dict):
        # unwrap a list of a single item in this context because it is
        # a mistake and we can infer their real meaning
        return parse_field_value(
            name,
            field[0],
            context,
        )
    # should be unreachable. Every code path to here tests types
    raise AssertionError(
        f"Unknown field {type(field)} type for {name}. Should be a string or 'object': \n {field} "
    )


def parse_field(name: str, definition, context: ParseContext) -> FieldFactory:
    assert name, name
    return FieldFactory(
        name,
        parse_field_value(name, definition, context),
        **context.line_num(definition),
    )


def parse_fields(fields: Dict, context: ParseContext) -> List[FieldFactory]:
    with context.change_current_parent_object(fields):
        # Should be unreachable because parser checks first.
        assert isinstance(
            fields, dict
        ), "Fields should be a dictionary (should not start with -) "

        return [
            parse_field(name, definition, context)
            for name, definition in fields.items()
            if name != "__line__"
        ]


def parse_friends(friends: List, context: ParseContext) -> List[Statement]:
    return parse_statement_list(friends, context)


def parse_count_expression(yaml_sobj: Dict, sobj_def: Dict, context: ParseContext):
    sobj_def["count_expr"] = parse_field_value("count", yaml_sobj["count"], context)


def include_macro(
    name: str, context: ParseContext, parent_macros=()
) -> Tuple[List[FieldFactory], List[TemplateLike]]:
    macro = context.macros.get(name)
    if not macro:
        raise exc.DataGenNameError(
            f"Cannot find macro named {name}", **context.line_num()
        )
    parsed_macro = parse_element(
        macro, "macro", {}, {"fields": Dict, "friends": List, "include": str}, context
    )
    if name in parent_macros:
        idx = parent_macros.index(name)
        raise exc.DataGenError(
            f"Macro `{name}` calls `{'` which calls `'.join(parent_macros[idx+1:])}` which calls `{name}`",
            **context.line_num(macro),
        )

    fields = []
    friends = []
    parse_inclusions(macro, fields, friends, context, parent_macros + (name,))
    fields.extend(parse_fields(parsed_macro.fields or {}, context))
    friends.extend(parse_friends(parsed_macro.friends or [], context))

    return _dedupe_field_list(fields), friends


def parse_inclusions(
    yaml_sobj: Dict,
    fields: List,
    friends: List,
    context: ParseContext,
    parent_macros=(),
) -> None:
    inclusions: Iterable[str] = [
        x.strip() for x in yaml_sobj.get("include", "").split(",")
    ]
    inclusions = filter(None, inclusions)
    for inclusion in inclusions:
        include_fields, include_friends = include_macro(
            inclusion, context, parent_macros
        )
        fields.extend(include_fields)
        friends.extend(include_friends)


# TODO: Use application object to write this instead
def check_identifier(name: T.Optional[str], source: dict, context: str):
    if not name:
        return
    badchars = set('."')
    if set(name).intersection(badchars):
        line_info = source["__line__"]
        warn(
            f"{name} is not a valid nickname.\n"
            f"{context} cannot contain '.' or '\"'."
            "Future versions of Snowfakery may disallow it.\n"
            f"{line_info.filename}:{line_info.line_num}",
            stacklevel=100,
        )


def _dedupe_field_list(fields: Sequence[FieldFactory]):
    return list({f.name: f for f in fields}.values())


def parse_object_template(yaml_sobj: Dict, context: ParseContext) -> ObjectTemplate:
    parsed_template: Any = parse_element(
        dct=yaml_sobj,
        element_type="object",
        mandatory_keys={},
        optional_keys={
            "fields": Dict,
            "friends": List,
            "include": str,
            "nickname": str,
            "just_once": bool,
            "for_each": dict,
            "count": (str, int, dict),
            "update_key": str,
        },
        context=context,
    )
    if not context.top_level and parsed_template.just_once:
        raise exc.DataGenSyntaxError(
            "just_once can only be used at the top level", **context.line_num()
        )

    assert yaml_sobj
    with context.change_current_parent_object(yaml_sobj):
        sobj_def = {}
        sobj_def["tablename"] = parsed_template.object
        check_identifier(parsed_template.object, yaml_sobj, "Object names")
        fields: List
        friends: List
        sobj_def["fields"] = fields = []
        sobj_def["friends"] = friends = []
        parse_inclusions(yaml_sobj, fields, friends, context)
        fields.extend(parse_fields(parsed_template.fields or {}, context))
        friends.extend(parse_friends(parsed_template.friends or [], context))
        fields[:] = _dedupe_field_list(fields)
        sobj_def["nickname"] = nickname = parsed_template.nickname
        check_identifier(nickname, yaml_sobj, "Nicknames")
        sobj_def["just_once"] = parsed_template.just_once or False
        sobj_def["update_key"] = parsed_template.update_key or None
        sobj_def["line_num"] = parsed_template.line_num.line_num
        sobj_def["filename"] = parsed_template.line_num.filename

        count_expr = yaml_sobj.get("count")

        if count_expr is not None:
            parse_count_expression(yaml_sobj, sobj_def, context)

        for_each_expr = yaml_sobj.get("for_each")
        if for_each_expr is not None:
            # it would be easy to support multiple parallel
            # for-eaches here and at one point the code did,
            # but the use-case wasn't obvious so it was removed
            # after 9bc296a7df. If we get to 2023 without
            # finding a use-case we can delete this comment.

            # this type should have been checked earlier
            assert isinstance(
                for_each_expr, dict
            ), "`for_each` must evaluate to a variable description"
            sobj_def["for_each_expr"] = parse_for_each_variable_definition(
                for_each_expr, context
            )

        new_template = ObjectTemplate(**sobj_def)
        context.register_template(new_template)
        return new_template


def parse_variable_definition(
    yaml_sobj: Dict, context: ParseContext
) -> VariableDefinition:
    parsed_template: Any = parse_element(
        yaml_sobj,
        "var",
        mandatory_keys={
            "value": (str, int, dict, list),
        },
        optional_keys={},
        context=context,
    )

    assert yaml_sobj
    with context.change_current_parent_object(yaml_sobj):
        sobj_def = {}
        sobj_def["varname"] = parsed_template.var
        var_def_expr = yaml_sobj.get("value")
        sobj_def["expression"] = parse_field_value("value", var_def_expr, context)
        sobj_def["line_num"] = parsed_template.line_num.line_num
        sobj_def["filename"] = parsed_template.line_num.filename
        new_def = VariableDefinition(**sobj_def)
        return new_def


def parse_for_each_variable_definition(
    yaml_sobj: Dict, context: ParseContext
) -> ForEachVariableDefinition:
    parsed_template: Any = parse_element(
        yaml_sobj,
        "var",
        optional_keys={},
        mandatory_keys={
            "value": (dict, str),
        },
        context=context,
    )

    assert yaml_sobj
    with context.change_current_parent_object(yaml_sobj):
        sobj_def = {}
        sobj_def["varname"] = parsed_template.var
        var_def_expr = yaml_sobj.get("value")

        sobj_def["expression"] = parse_field_value("value", var_def_expr, context)
        sobj_def["line_num"] = parsed_template.line_num.line_num
        sobj_def["filename"] = parsed_template.line_num.filename
        new_def = ForEachVariableDefinition(**sobj_def)
        return new_def


def parse_statement_list(
    statements: List[dict], context: ParseContext
) -> List[Statement]:
    parsed_statements = []
    for obj in statements:
        assert isinstance(obj, dict), obj
        if obj.get("object"):
            object_template = parse_object_template(obj, context)
            parsed_statements.append(object_template)
        elif obj.get("var"):
            variable_definition = parse_variable_definition(obj, context)
            parsed_statements.append(variable_definition)
        else:
            keys = [key for key in obj.keys() if not key.startswith("_")]
            raise exc.DataGenSyntaxError(
                f"This statement cannot be parsed: {keys}", **context.line_num(obj)
            )
    return parsed_statements


def yaml_safe_load_with_line_numbers(
    filestream: IO[str], filename: str
) -> Tuple[object, Dict]:
    loader = yaml.SafeLoader(filestream)
    line_numbers: Dict[Any, LineTracker] = {}

    def compose_node(parent, index):
        # the line number where the previous token has ended (plus empty lines)
        line = loader.line
        node = Composer.compose_node(loader, parent, index)
        node.__line__ = line + 1
        return node

    def construct_mapping(node, deep=False):
        mapping = SafeConstructor.construct_mapping(loader, node, deep=deep)
        mapping["__line__"] = LineTracker(filename, node.__line__)
        return mapping

    def construct_scalar(node):
        scalar = SafeConstructor.construct_scalar(loader, node)
        key = id(scalar)
        if not line_numbers.get(key):
            line_numbers[key] = LineTracker(filename, node.__line__)
        else:
            line_numbers[key] = SHARED_OBJECT
        return scalar

    loader.compose_node = compose_node  # type: ignore
    loader.construct_mapping = construct_mapping  # type: ignore
    loader.construct_scalar = construct_scalar  # type: ignore
    return loader.get_single_data(), line_numbers


class DictValuesAsAttrs:
    pass


def parse_element(
    dct: Dict,
    element_type: str,
    mandatory_keys: Dict,
    optional_keys: Dict,
    context: ParseContext,
) -> Any:
    expected_keys = {
        **mandatory_keys,
        **optional_keys,
        "__line__": LineTracker,
        element_type: str,
    }
    rc_obj: Any = DictValuesAsAttrs()
    rc_obj.line_num = dct["__line__"]
    with context.change_current_parent_object(dct):
        for key in dct:
            key_definition = expected_keys.get(key)
            if not key_definition:
                raise exc.DataGenSyntaxError(
                    f"Unexpected key: {key}", **context.line_num(key)
                )
            else:
                value = dct[key]
                if not isinstance(value, key_definition):
                    raise exc.DataGenSyntaxError(
                        f"Expected `{key}` to be of type {key_definition} instead of `{type(value).__name__}`.",
                        **context.line_num(dct),
                    )
                else:
                    setattr(rc_obj, key, value)

        missing_keys = set(mandatory_keys) - set(dct.keys())
        if missing_keys:
            raise exc.DataGenError(
                f"Expected to see `{missing_keys}` in `{element_type}``.",
                **context.line_num(dct),
            )
        defaulted_keys = set(optional_keys) - set(dct.keys())
        for key in defaulted_keys:
            setattr(rc_obj, key, None)

        return rc_obj


def relpath_from_inclusion_element(
    inclusion: Dict, context: ParseContext
) -> Tuple[Path, LineTracker]:
    # should be a two-element dict: {'include_file': 'foo.yml', "__line__": 5}
    inclusion_parsed: Any = parse_element(inclusion, "include_file", {}, {}, context)
    relpath = inclusion_parsed.include_file
    linenum = inclusion_parsed.line_num or LineTracker("unknown", -1)
    assert not relpath.startswith("/")  # only relative paths
    return Path(relpath), linenum


def parse_included_file(
    parent_path: Path, inclusion: Dict, context: ParseContext
) -> List[Dict]:
    relpath, linenum = relpath_from_inclusion_element(inclusion, context)
    inclusion_path = parent_path.parent / relpath
    # someday add a check that we don't go outside of the project dir
    if not inclusion_path.exists():
        raise exc.DataGenError(
            f"Cannot load include file {inclusion_path}", **linenum._asdict()
        )
    with inclusion_path.open() as f:
        incl_objects = parse_file(f, context)
        return incl_objects


def parse_included_files(path: Path, data: List, context: ParseContext):
    file_inclusions = [obj for obj in data if obj.get("include_file")]

    templates = []
    for fi in file_inclusions:
        templates.extend(parse_included_file(path, fi, context))
    return templates


collection_rules = {
    "option": "option",
    "include_file": "include_file",
    "macro": "macro",
    "plugin": "plugin",
    "object": "statement",
    "var": "statement",
    "snowfakery_version": "snowfakery_version",
}


def categorize_top_level_objects(data: List, context: ParseContext):
    """Look at all of the top-level declarations and categorize them"""
    top_level_collections: Dict = {
        "snowfakery_version": [],
        "option": [],
        "include_file": [],
        "macro": [],
        "plugin": [],
        "statement": [],  # objects with side-effects
    }
    assert isinstance(data, list)
    for obj in data:
        if not isinstance(obj, dict):
            raise exc.DataGenSyntaxError(
                f"Top level elements in a data generation template should all be dictionaries, not {obj}",
            )
        obj_category = None
        for declaration, category in collection_rules.items():
            typ = obj.get(declaration)
            if typ:
                if obj_category:
                    raise exc.DataGenError(
                        f"Top level element seems to match two name patterns: {declaration, obj_category}",
                        **context.line_num(obj),
                    )
                obj_category = category
        if obj_category:
            top_level_collections[obj_category].append(obj)
        else:
            raise exc.DataGenError(
                f"Unknown object type {obj}", **context.line_num(obj)
            )
    return top_level_collections


def parse_top_level_elements(path: Path, data: List, context: ParseContext):
    top_level_objects = categorize_top_level_objects(data, context)
    statements: List[ObjectTemplate] = []
    statements.extend(parse_included_files(path, data, context))
    context.options.extend(top_level_objects["option"])
    context.macros.update({obj["macro"]: obj for obj in top_level_objects["macro"]})
    plugin_specs = [
        (obj["plugin"], obj["__line__"]) for obj in top_level_objects["plugin"]
    ]
    plugin_near_recipe = path.parent / "plugins"
    context.plugins.extend(
        resolve_plugins(plugin_specs, search_paths=[plugin_near_recipe])
    )
    context.version = parse_version(top_level_objects["snowfakery_version"], context)
    statements.extend(top_level_objects["statement"])
    for pluginbase, plugin in context.plugins:
        if pluginbase == ParserMacroPlugin:
            context.parser_macros_plugins[plugin.__name__] = plugin()
    return statements


def parse_version(version_declarations: T.List[T.Dict], context: ParseContext):
    if version_declarations:
        base_version = version_declarations[0]["snowfakery_version"]
        mismatched_versions = [
            obj
            for obj in version_declarations
            if obj["snowfakery_version"] != base_version
        ]
        if mismatched_versions:
            with context.change_current_parent_object(version_declarations[1]):
                raise exc.DataGenSyntaxError(
                    "Cannot have multiple conflicting versions in the same recipe: ",
                    **context.line_num(),
                )
        if base_version not in (2, 3):
            with context.change_current_parent_object(version_declarations[0]):
                raise exc.DataGenSyntaxError(
                    "Version must be 2 or 3: ",
                    **context.line_num(),
                )
        return base_version


def parse_file(stream: IO[str], context: ParseContext) -> List[Dict]:
    stream_name = getattr(stream, "name", None)
    if stream_name:
        path = Path(fsdecode(stream.name)).absolute()
    else:
        path = Path("<stream>")
    try:
        data, line_numbers = yaml_safe_load_with_line_numbers(stream, str(path))
    except YAMLError as y:
        raise exc.DataGenYamlSyntaxError(
            str(y),
            str(path),
            y.problem_mark.line + 1,
        )
    context.line_numbers.update(line_numbers)

    if not isinstance(data, list):
        raise exc.DataGenSyntaxError(
            "Recipe file should be a list (use '-' on top-level lines)",
            stream_name,
            1,
        )

    statements = parse_top_level_elements(path, data, context)

    return statements


def build_update_recipe(
    statements: List[Statement],
    tables: dict,
    update_input_file: FileLike = None,
    passthrough_fields: T.Sequence = (),
) -> List[Statement]:
    class DataSourceValue(SimpleValue):
        def __init__(self):
            self.datasource = CSVDatasetLinearIterator(update_input_file, False)

        def render(self, *args, **kwargs):
            return self.datasource

    error_message = "Update recipes should have a single object declaration."
    if len(statements) != 1:
        raise exc.DataGenSyntaxError(error_message)
    template = statements[0]
    if not isinstance(template, ObjectTemplate):
        raise exc.DataGenSyntaxError(
            error_message,
            filename=template.filename,
            line_num=template.line_num,
        )

    if template.count_expr:
        raise exc.DataGenSyntaxError(
            "Update templates should have no 'count'",
            filename=template.filename,
            line_num=template.line_num,
        )

    template.for_each_expr = ForEachVariableDefinition(
        "", -1, "input", DataSourceValue()
    )

    def _make_passthrough_attribute_for_update_recipe(attrname):
        # the magic numbers are fake line-numbers. They should be irrelevant
        # but perhaps they will be useful in debugging some future problem
        field_def = SimpleValue("${{input.%s}}" % attrname, "", -3)
        new_field = FieldFactory(attrname, field_def, "", -4)
        tables[template.name].fields[attrname] = new_field
        return new_field

    template.fields.extend(
        _make_passthrough_attribute_for_update_recipe(attr)
        for attr in passthrough_fields
    )

    return [template]


def parse_recipe(
    stream: IO[str],
    update_input_file: FileLike = None,
    update_passthrough_fields: T.Sequence[str] = (),
) -> ParseResult:
    context = ParseContext()
    objects = parse_file(stream, context)  # parse the yaml without semantics
    statements = parse_statement_list(objects, context)
    tables = context.table_infos
    tables = {
        name: value
        for name, value in context.table_infos.items()
        if not name.startswith("__")
    }
    if update_input_file:
        statements = build_update_recipe(
            statements, tables, update_input_file, update_passthrough_fields
        )

    return ParseResult(
        context.options,
        tables,
        statements,
        plugins=context.plugins,
        version=context.version,
        random_references=context.random_references,
    )
